# -*- coding: utf-8 -*-

'''
To test the pdf conversion and converting quality, the idea is to convert generated docx to pdf,
then check the image similarity between source pdf page and converted pdf page. Considering the 
converting quality from docx to pdf, a Windows-based command line tool `OfficeToPDF` is used, in
addition, an installation of Microsoft Word is required.

To leverage the benefit of Github Action, the testing process is divided into three parts:
  1. Convert sample pdf to docx with `pdf2docx`.
  2. Convert generated docx to pdf for comparing. 
  3. Convert page to image and compare similarity with `python-opencv`.

Test scripts on Part One and Three are applied with two test class respectively in this module,
so they could be run seperately with pytest command, e.g.

- pytest -vs --no-header test.py::TestConversion for Part One
- pytest -vs --no-header test.py::TestQuality for Part Three

Links on MS Word to PDF conversion:
  - https://github.com/cognidox/OfficeToPDF/releases
  - https://github.com/AndyCyberSec/pylovepdf
  - https://www.e-iceblue.com/Tutorials/Java/Spire.Doc-for-Java/Program-Guide/Conversion/Convert-Word-to-PDF-in-Java.html
'''

import os
import numpy as np
import cv2 as cv
import fitz
from pdf2docx import Converter, parse


script_path = os.path.abspath(__file__) # current script path
test_dir = os.path.dirname(script_path)
sample_path = os.path.join(test_dir, 'samples')
output_path = os.path.join(test_dir, 'outputs')


def get_page_similarity(page_a, page_b, diff_img_filename='diff.png'):
    '''Calculate page similarity index: [0, 1].'''
    # page to opencv image
    image_a = get_page_image(page_a)
    image_b = get_page_image(page_b)

    # resize if different shape
    if image_a.shape != image_b.shape:
        rows, cols = image_a.shape[:2]
        image_b = cv.resize(image_b, (cols, rows))
    
    # write different image
    if diff_img_filename:
        diff = cv.subtract(image_a, image_b)
        cv.imwrite(diff_img_filename, diff)
    
    return get_mssism(image_a, image_b)


def get_page_image(pdf_page):
    '''Convert fitz page to opencv image.'''
    img_byte = pdf_page.get_pixmap(clip=pdf_page.rect).tobytes()
    img = np.frombuffer(img_byte, np.uint8)
    return cv.imdecode(img, cv.IMREAD_COLOR)


def get_mssism(i1, i2, kernel=(15,15)):
    '''Calculate mean Structural Similarity Index (SSIM).
    https://docs.opencv.org/4.x/d5/dc4/tutorial_video_input_psnr_ssim.html
    '''
    C1 = 6.5025
    C2 = 58.5225
    # INITS
    I1 = np.float32(i1) # cannot calculate on one byte large values
    I2 = np.float32(i2)
    I2_2 = I2 * I2 # I2^2
    I1_2 = I1 * I1 # I1^2
    I1_I2 = I1 * I2 # I1 * I2
    # END INITS
    # PRELIMINARY COMPUTING
    mu1 = cv.GaussianBlur(I1, kernel, 1.5)
    mu2 = cv.GaussianBlur(I2, kernel, 1.5)
    mu1_2 = mu1 * mu1
    mu2_2 = mu2 * mu2
    mu1_mu2 = mu1 * mu2
    sigma1_2 = cv.GaussianBlur(I1_2, kernel, 1.5)
    sigma1_2 -= mu1_2
    sigma2_2 = cv.GaussianBlur(I2_2, kernel, 1.5)
    sigma2_2 -= mu2_2
    sigma12 = cv.GaussianBlur(I1_I2, kernel, 1.5)
    sigma12 -= mu1_mu2
    t1 = 2 * mu1_mu2 + C1
    t2 = 2 * sigma12 + C2
    t3 = t1 * t2                    # t3 = ((2*mu1_mu2 + C1).*(2*sigma12 + C2))
    t1 = mu1_2 + mu2_2 + C1
    t2 = sigma1_2 + sigma2_2 + C2
    t1 = t1 * t2                    # t1 =((mu1_2 + mu2_2 + C1).*(sigma1_2 + sigma2_2 + C2))
    ssim_map = cv.divide(t3, t1)    # ssim_map =  t3./t1;
    mssim = cv.mean(ssim_map)       # mssim = average of ssim map
    return np.mean(mssim[0:3])



class TestConversion:
    '''Test the converting process.'''

    def setup(self):
        '''create output path if not exist.'''
        if not os.path.exists(output_path): os.mkdir(output_path)
    

    def convert(self, filename):
        '''Convert PDF file from sample path to output path.'''
        source_pdf_file = os.path.join(sample_path, f'{filename}.pdf')
        docx_file = os.path.join(output_path, f'{filename}.docx')
        cv = Converter(source_pdf_file)        
        cv.convert(docx_file)
        cv.close()    

    # ------------------------------------------
    # layout: section
    # ------------------------------------------
    def test_section(self):
        '''test page layout: section and column.'''
        self.convert('demo-section')    

    def test_section_spacing(self):
        '''test page layout: section vertical position.'''
        self.convert('demo-section-spacing')

    # ------------------------------------------
    # text styles
    # ------------------------------------------
    def test_blank_file(self):
        '''test blank file without any texts or images.'''
        self.convert('demo-blank')

    def test_text_format(self):
        '''test text format, e.g. highlight, underline, strike-through.'''
        self.convert('demo-text')

    def test_text_alignment(self):
        '''test text alignment.'''
        self.convert('demo-text-alignment')    
    
    def test_unnamed_fonts(self):
        '''test unnamed fonts which destroys span bbox, and accordingly line/block layout.'''
        self.convert('demo-text-unnamed-fonts')

    def test_text_scaling(self):
        '''test font size. In this case, the font size is set precisely with character scaling.'''
        self.convert('demo-text-scaling')
    
    def test_text_hidden(self):
        '''test hidden text, which is ignore by default.'''
        self.convert('demo-text-hidden')

    # ------------------------------------------
    # image styles
    # ------------------------------------------
    def test_image(self):
        '''test inline-image.'''
        self.convert('demo-image')

    def test_vector_graphic(self):
        '''test vector graphic.'''
        self.convert('demo-image-vector-graphic')

    def test_image_cmyk(self):
        '''test image in CMYK color-space.'''
        self.convert('demo-image-cmyk')

    def test_image_transparent(self):
        '''test transparent images.'''
        self.convert('demo-image-transparent')
    
    def test_image_rotation(self):
        '''test rotating image due to pdf page rotation.'''
        self.convert('demo-image-rotation')

    def test_image_overlap(self):
        '''test images with both intersection and page rotation.'''
        self.convert('demo-image-overlap')


    # ------------------------------------------
    # table styles
    # ------------------------------------------
    def test_table_bottom(self):
        '''page break due to table at the end of page.'''
        self.convert('demo-table-bottom')

    def test_table_format(self):
        '''test table format, e.g. 
            - border and shading style
            - vertical cell
            - merged cell
            - text format in cell
        '''
        self.convert('demo-table')

    def test_stream_table(self):
        '''test stream structure and shading.'''
        self.convert('demo-table-stream')

    def test_table_shading(self):
        '''test simulating shape with shading cell.'''
        self.convert('demo-table-shading')
    
    def test_table_shading_highlight(self):
        '''test distinguishing table shading and highlight.'''
        self.convert('demo-table-shading-highlight')

    def test_lattice_table(self):
        '''test lattice table with very close text underlines to table borders.'''
        self.convert('demo-table-close-underline')

    def test_lattice_table_invoice(self):
        '''test invoice sample file with lattice table, vector graphic.'''
        self.convert('demo-table-lattice')

    def test_lattice_cell(self):
        '''test generating stream borders for lattice table cell.'''
        self.convert('demo-table-lattice-one-cell')

    def test_table_border_style(self):
        '''test border style, e.g. width, color.'''
        self.convert('demo-table-border-style')

    def test_table_align_borders(self):
        '''aligning stream table borders to simplify table structure.'''
        self.convert('demo-table-align-borders')

    def test_nested_table(self):
        '''test nested tables.'''
        self.convert('demo-table-nested')

    def test_path_transformation(self):
        '''test path transformation. In this case, the (0,0) origin is out of the page.'''
        self.convert('demo-path-transformation')


    # ------------------------------------------
    # table contents
    # ------------------------------------------
    def test_extracting_table(self):
        '''test extracting contents from table.'''
        filename = 'demo-table'
        pdf_file = os.path.join(sample_path, f'{filename}.pdf')
        tables = Converter(pdf_file).extract_tables(end=1)
        print(tables)

        # compare the last table
        table = [[col.strip() if col else col for col in row] for row in tables[-1]]
        sample = [
            ['Input', None, None, None, None, None],
            ['Description A', 'mm', '30.34', '35.30', '19.30', '80.21'],
            ['Description B', '1.00', '5.95', '6.16', '16.48', '48.81'],
            ['Description C', '1.00', '0.98', '0.94', '1.03', '0.32'],
            ['Description D', 'kg', '0.84', '0.53', '0.52', '0.33'],
            ['Description E', '1.00', '0.15', None, None, None],
            ['Description F', '1.00', '0.86', '0.37', '0.78', '0.01']
        ]
        assert table==sample

    
    # ------------------------------------------
    # command line arguments
    # ------------------------------------------
    def test_multi_pages(self):
        '''test converting pdf with multi-pages.'''
        filename = 'demo'
        pdf_file = os.path.join(sample_path, f'{filename}.pdf')
        docx_file = os.path.join(output_path, f'{filename}.docx')    
        parse(pdf_file, docx_file, start=1, end=5)

        # check file        
        assert os.path.isfile(docx_file)
    


class TestQuality:
    '''Check the quality of converted docx. 
    Note the docx files must be converted to PDF files in advance.
    '''

    INDEX_MAP = {
        'demo-blank.pdf': 1.0,
        'demo-image-cmyk.pdf': 0.90,
        'demo-image-transparent.pdf': 0.90,
        'demo-image-vector-graphic.pdf': 0.89,
        'demo-image.pdf': 0.90,
        'demo-image-rotation.pdf': 0.90,
        'demo-image-overlap.pdf': 0.90,
        'demo-path-transformation.pdf': 0.90,
        'demo-section-spacing.pdf': 0.90,
        'demo-section.pdf': 0.70,
        'demo-table-align-borders.pdf': 0.49,
        'demo-table-border-style.pdf': 0.90,
        'demo-table-bottom.pdf': 0.90,
        'demo-table-close-underline.pdf': 0.58,
        'demo-table-lattice-one-cell.pdf': 0.79,
        'demo-table-lattice.pdf': 0.75,
        'demo-table-nested.pdf': 0.84,
        'demo-table-shading-highlight.pdf': 0.55,
        'demo-table-shading.pdf': 0.80,
        'demo-table-stream.pdf': 0.55,
        'demo-table.pdf': 0.90,
        'demo-text-alignment.pdf': 0.90,
        'demo-text-scaling.pdf': 0.80,
        'demo-text-unnamed-fonts.pdf': 0.80,
        'demo-text-hidden.pdf': 0.90,
        'demo-text.pdf': 0.80
    }

    def setup(self):
        '''create output path if not exist.'''
        if not os.path.exists(output_path): os.mkdir(output_path)


    def test_quality(self):
        '''Convert page to image and compare similarity.'''
        for filename in os.listdir(output_path):
            if not filename.endswith('pdf'): continue

            source_pdf_file = os.path.join(sample_path, filename)
            target_pdf_file = os.path.join(output_path, filename)

            # open pdf    
            source_pdf = fitz.open(source_pdf_file)
            target_pdf = fitz.open(target_pdf_file)

            # compare page count
            if len(source_pdf)>1: continue # one page sample only
            assert len(target_pdf)==1, f"\nThe page count of {filename} is incorrect."

            # compare the first page
            diff_png = os.path.join(output_path, f'{filename[:-4]}.png')
            sidx = get_page_similarity(target_pdf[0], source_pdf[0], diff_png)
            threshold = TestQuality.INDEX_MAP.get(filename, 0.10)
            print(f'Checking {filename}: {sidx} v.s. {threshold}')
            assert sidx>=threshold, 'Significant difference might exist since similarity index is lower than threshold.'



# -*- coding: utf-8 -*-

'''A group of instances, e.g. Blocks, Lines, Spans, Shapes.
'''

import fitz
from .Element import Element
from .share import (IText, TextDirection)
from .algorithm import (solve_rects_intersection, graph_bfs)


class BaseCollection:
    '''Base collection representing a list of instances.'''
    def __init__(self, instances:list=None, parent=None):
        '''Init collection from a list of instances.'''
        self._parent = parent
        self._instances = []
        self.extend(instances or []) # Note to exclude empty instance by default

    def __getitem__(self, idx):
        try:
            instances = self._instances[idx]
        except IndexError:
            msg = f'Collection index {idx} out of range.'
            raise IndexError(msg)
        else:
            return instances

    def __iter__(self): return (instance for instance in self._instances)

    def __len__(self): return len(self._instances)

    @property
    def parent(self): return self._parent


    @property
    def bbox(self):
        '''bbox of combined collection.'''
        rect = fitz.Rect()
        for instance in self._instances:
            rect |= instance.bbox
        return fitz.Rect([round(x,1) for x in rect]) # NOTE: round to avoid digital error


    def append(self, instance): 
        if not instance: return
        self._instances.append(instance)


    def extend(self, instances:list): 
        if not instances: return
        for instance in instances: self.append(instance)


    def reset(self, instances:list=None):
        """Reset instances list.

        Args:
            instances (list, optional): reset to target instances. Defaults to None.

        Returns:
            BaseCollection: self
        """
        self._instances = []
        self.extend(instances or [])
        return self


    def store(self):
        '''Store attributes in json format.'''
        return [ instance.store() for instance in self._instances ]


    def restore(self, *args, **kwargs):
        '''Construct Collection from a list of dict.'''
        raise NotImplementedError


class Collection(BaseCollection, IText):
    '''Collection of instance focusing on grouping and sorting elements.'''
    
    @property
    def text_direction(self):
        '''Get text direction. All instances must have same text direction.''' 
        res = set(instance.text_direction for instance in self._instances)
        return list(res)[0] if len(res)==1 else TextDirection.MIX 


    def group(self, fun):
        """Group instances according to user defined criterion.

        Args:
            fun (function): with 2 arguments representing 2 instances (Element) and return bool.

        Returns:
            list: a list of grouped ``Collection`` instances.
        
        Examples 1::

            # group instances intersected with each other
            fun = lambda a,b: a.bbox & b.bbox
        
        Examples 2::

            # group instances aligned horizontally
            fun = lambda a,b: a.horizontally_aligned_with(b)
        
        .. note::
            It's equal to a GRAPH searching problem, build adjacent list, and then search graph
            to find all connected components.
        """
        # build adjacent list:
        # the i-th item is a set of indexes, which connected to the i-th instance.
        # NOTE: O(n^2) method, but it's acceptable (~0.2s) when n<1000 which is satisfied by page blocks
        num = len(self._instances)
        index_groups = [set() for i in range(num)] # type: list[set]        
        for i, instance in enumerate(self._instances):
            # connections of current instance to all instances after it
            for j in range(i+1, num):
                if fun(instance, self._instances[j]):
                    index_groups[i].add(j)
                    index_groups[j].add(i)

        # search graph -> grouped index of instance
        groups = graph_bfs(index_groups)
        groups = [self.__class__([self._instances[i] for i in group]) for group in groups]
        return groups

    
    def group_by_connectivity(self, dx:float, dy:float):
        """Collect connected instances into same group.

        Args:
            dx (float): x-tolerances to define connectivity
            dy (float): y-tolerances to define connectivity

        Returns:
            list: a list of grouped ``Collection`` instances.
        
        .. note::
            * It's equal to a GRAPH traversing problem, which the critical point in 
              building the adjacent list, especially a large number of vertex (paths).

            * Checking intersections between paths is actually a Rectangle-Intersection 
              problem, studied already in many literatures.
        """
        # build the graph -> adjacent list:
        # the i-th item is a set of indexes, which connected to the i-th instance
        num = len(self._instances)
        index_groups = [set() for _ in range(num)] # type: list[set]

        # solve rectangle intersection problem
        i_rect_x, i = [], 0
        d_rect = (-dx, -dy, dx, dy)
        for rect in self._instances:
            points = [a+b for a,b in zip(rect.bbox, d_rect)] # consider tolerance
            i_rect_x.append((i,   points, points[0]))
            i_rect_x.append((i+1, points, points[2]))
            i += 2
        i_rect_x.sort(key=lambda item: item[-1])
        solve_rects_intersection(i_rect_x, 2*num, index_groups)

        # search graph -> grouped index of instance
        groups = graph_bfs(index_groups)
        groups = [self.__class__([self._instances[i] for i in group]) for group in groups]
        return groups
    
    
    def group_by_columns(self, factor:float=0.0, sorted:bool=True, text_direction:bool=False):
        '''Group elements into columns based on the bbox.'''
        # split in columns
        fun = lambda a,b: a.vertically_align_with(b, factor=factor, text_direction=text_direction)
        groups = self.group(fun)
        
        # increase in x-direction if sort
        if sorted: 
            idx = 3 if text_direction and self.is_vertical_text else 0
            groups.sort(key=lambda group: group.bbox[idx])

        return groups


    def group_by_rows(self, factor:float=0.0, sorted:bool=True, text_direction:bool=False):
        '''Group elements into rows based on the bbox.'''
        # split in rows
        fun = lambda a,b: a.horizontally_align_with(b, factor=factor, text_direction=text_direction)
        groups = self.group(fun)

        # increase in y-direction if sort
        if sorted: 
            idx = 0 if text_direction and self.is_vertical_text else 1
            groups.sort(key=lambda group: group.bbox[idx])

        return groups


    def group_by_physical_rows(self, sorted:bool=False, text_direction:bool=False):
        '''Group lines into physical rows.'''
        fun = lambda a,b: a.in_same_row(b)
        groups = self.group(fun)

        # increase in y-direction if sort
        if sorted: 
            idx = 0 if text_direction and self.is_vertical_text else 1
            groups.sort(key=lambda group: group.bbox[idx])

        return groups


    def sort_in_reading_order(self):
        '''Sort collection instances in reading order (considering text direction), e.g.
            for normal reading direction: from top to bottom, from left to right.
        '''
        if self.is_horizontal_text:
            self._instances.sort(key=lambda e: (e.bbox.y0, e.bbox.x0, e.bbox.x1))
        else:
            self._instances.sort(key=lambda e: (e.bbox.x0, e.bbox.y1, e.bbox.y0))
        return self


    def sort_in_line_order(self):
        '''Sort collection instances in a physical with text direction considered, e.g.
            for normal reading direction: from left to right.
        '''
        if not self.is_vertical_text:
            self._instances.sort(key=lambda e: (e.bbox.x0, e.bbox.y0, e.bbox.x1))
        else:
            self._instances.sort(key=lambda e: (e.bbox.y1, e.bbox.x0, e.bbox.y0))
        return self


    def sort_in_reading_order_plus(self):
        '''Sort instances in reading order, especially for instances in same row. Taking 
        natural reading direction for example: reading order for rows, from left to right 
        for instances in row. In the following example, A comes before B::

                         +-----------+
            +---------+  |           |
            |   A     |  |     B     |
            +---------+  +-----------+
        
        Steps:

            * Sort elements in reading order, i.e. from top to bottom, from left to right.
            * Group elements in row.
            * Sort elements in row: from left to right.
        '''
        instances = []
        for row in self.group_by_physical_rows(sorted=True, text_direction=True):
            row.sort_in_line_order()
            instances.extend(row)        
        self.reset(instances)



class ElementCollection(Collection):
    '''Collection of ``Element`` instances.'''

    def _update_bbox(self, e:Element):
        '''Update parent bbox.'''
        if not self._parent is None: # Note: `if self._parent` does not work here
            self._parent.union_bbox(e)


    def append(self, e:Element):
        """Append an instance, update parent's bbox accordingly and set the parent of the added instance.

        Args:
            e (Element): instance to append.
        """
        if not e: return
        self._instances.append(e)
        self._update_bbox(e)

        # set parent
        if not self._parent is None: e.parent = self._parent 


    def insert(self, nth:int, e:Element):
        """Insert a Element and update parent's bbox accordingly.

        Args:
            nth (int): the position to insert.
            e (Element): the instance to insert.
        """        
        if not e: return
        self._instances.insert(nth, e)
        self._update_bbox(e)
        e.parent = self._parent # set parent

    
    def pop(self, nth:int):
        """Delete the ``nth`` instance.

        Args:
            nth (int): the position to remove.

        Returns:
            Collection: the removed instance.
        """        
        return self._instances.pop(nth)


    def is_flow_layout(self, line_separate_threshold:float, cell_layout=False):
        '''Whether contained elements are in flow layout or not.'''
        # float layout if vertical text but not cell layout, since vertical text 
        # will be simulated with stream table
        if not cell_layout and self.is_vertical_text:
            return False

        # flow layout if single column only
        if len(self)<=1: return True
        if len(self.group_by_columns())>1: return False

        # group in physical row and check distance between lines
        idx0, idx1 = (0, 2) if self.is_horizontal_text else (3, 1)
        for row in self.group_by_physical_rows(text_direction=True):
            for i in range(1, len(row)):
                dis = abs(row[i].bbox[idx0]-row[i-1].bbox[idx1])
                if dis >= line_separate_threshold: return False

        return True


    def contained_in_bbox(self, bbox):
        '''Filter instances contained in target bbox.

        Args:
            bbox  (fitz.Rect): target boundary box.
        '''
        instances = list(filter(
            lambda e: bbox.contains(e.bbox), self._instances))
        return self.__class__(instances)


    def split_with_intersection(self, bbox:fitz.Rect, threshold:float=1e-3):
        """Split instances into two groups: one intersects with ``bbox``, the other not.

        Args:
            bbox (fitz.Rect): target rect box.
            threshold (float): It's intersected when the overlap rate exceeds this threshold. Defaults to 0.

        Returns:
            tuple: two group in original class type.
        """
        intersections, no_intersections = [], []
        for instance in self._instances:
            # A contains B => A & B = B
            intersection = instance.bbox & bbox
            factor = round(intersection.get_area()/instance.bbox.get_area(), 2)

            if factor >= threshold:
                intersections.append(instance)
            else:
                no_intersections.append(instance)
        return self.__class__(intersections), self.__class__(no_intersections)


# -*- coding: utf-8 -*-

'''Collection of :py:class:`~pdf2docx.layout.Section` instances.
'''

from docx.enum.section import WD_SECTION
from docx.shared import Pt
from ..common.Collection import BaseCollection
from ..common.docx import reset_paragraph_format
from .Section import Section
from ..common import constants


class Sections(BaseCollection):

    def restore(self, raws:list):
        """Restore sections from source dicts."""        
        self.reset()
        for raw in raws:
            section = Section().restore(raw)
            self.append(section)
        return self
    

    def parse(self, **settings):
        '''Parse layout under section level.'''
        for section in self: section.parse(**settings)
        return self


    def make_docx(self, doc):
        '''Create sections in docx.'''        
        if not self: return

        # mark paragraph index before creating current page
        n = len(doc.paragraphs)

        def create_dummy_paragraph_for_section(section):
            p = doc.add_paragraph()
            line_height = min(section.before_space, 11)
            pf = reset_paragraph_format(p, line_spacing=Pt(line_height))
            pf.space_after = Pt(section.before_space-line_height)

        # ---------------------------------------------------
        # first section
        # ---------------------------------------------------
        # vertical position: add dummy paragraph only if before space is required
        section = self[0]
        if section.before_space > constants.MINOR_DIST:
            create_dummy_paragraph_for_section(section)
        
        # create first section
        if section.num_cols==2: 
            doc.add_section(WD_SECTION.CONTINUOUS)
        section.make_docx(doc)

        # ---------------------------------------------------
        # more sections
        # ---------------------------------------------------
        for section in self[1:]:
            # create new section symbol
            doc.add_section(WD_SECTION.CONTINUOUS)

            # set after space of last paragraph to define the vertical
            # position of current section
            # NOTE: the after space doesn't work if last paragraph is 
            # image only (without any text). In this case, set after
            # space for the section break.
            p = doc.paragraphs[-2] # -1 is the section break
            if not p.text.strip() and 'graphicData' in p._p.xml:
                p = doc.paragraphs[-1]
            pf = p.paragraph_format
            pf.space_after = Pt(section.before_space)
            
            # section content
            section.make_docx(doc)

        # ---------------------------------------------------
        # create floating images
        # ---------------------------------------------------
        # lazy: assign all float images to first paragraph of current page
        for image in self.parent.float_images:
            image.make_docx(doc.paragraphs[n])


    def plot(self, page):
        '''Plot all section blocks for debug purpose.'''
        for section in self: 
            for column in section:
                column.plot(page, stroke=(1,1,0), width=1.5) # column bbox
                column.blocks.plot(page) # blocks


# -*- coding: utf-8 -*-

'''
Text Line objects based on PDF raw dict extracted with ``PyMuPDF``.

Data structure of line in text block referring to this
`link <https://pymupdf.readthedocs.io/en/latest/textpage.html>`_::

    {
        'bbox': (x0,y0,x1,y1),
        'wmode': m,
        'dir': [x,y],
        'spans': [ spans ]
    }
'''

from fitz import Point
try:
    # Python <= 3.9
    from collections import Iterable
except ImportError:
    # Python > 3.9
    from collections.abc import Iterable
from .TextSpan import TextSpan
from ..common.Element import Element
from ..common.share import TextDirection
from .Spans import Spans
from ..image.ImageSpan import ImageSpan


class Line(Element):
    '''Object representing a line in text block.'''
    def __init__(self, raw:dict=None):
        if raw is None: raw = {}

        # writing mode
        self.wmode = raw.get('wmode', 0) 

        # update writing direction to rotated page CS
        if 'dir' in raw:
            self.dir = list(Point(raw['dir'])*Line.pure_rotation_matrix())
        else:
            self.dir = [1.0, 0.0] # left -> right by default

        # line break
        self.line_break = raw.get('line_break', 0) # don't break line by default
        self.tab_stop = raw.get('tab_stop', 0) # no TAB stop before the line by default
        
        # remove key 'bbox' since it is calculated from contained spans
        if 'bbox' in raw: raw.pop('bbox') 
        super().__init__(raw)

        # collect spans
        self.spans = Spans(parent=self).restore(raw.get('spans', []))        

    
    @property
    def text(self):
        '''Joining span text. Note image is translated to a placeholder ``<image>``.'''
        spans_text = [span.text for span in self.spans]
        return ''.join(spans_text)


    @property
    def raw_text(self):
        '''Joining span text with image ignored.'''
        spans_text = [span.text for span in self.spans if isinstance(span, TextSpan)]
        return ''.join(spans_text)


    @property
    def white_space_only(self):
        '''If this line contains only white space or not. If True, this line is safe to be removed.'''
        for span in self.spans:
            if not isinstance(span, TextSpan): return False
            if span.text.strip(): return False
        return True


    @property
    def image_spans(self):
        '''Get image spans in this Line.'''
        return list(filter(
            lambda span: isinstance(span, ImageSpan), self.spans
        ))


    @property
    def text_direction(self):
        '''Get text direction. Consider ``LEFT_RIGHT`` and ``LEFT_RIGHT`` only.

        Returns:
            TextDirection: Text direction of this line.
        '''
        if self.dir[0] == 1.0:
            return TextDirection.LEFT_RIGHT
        elif self.dir[1] == -1.0:
            return TextDirection.BOTTOM_TOP
        else:
            return TextDirection.IGNORE


    def strip(self):
        '''Remove redundant blanks at the begin/end span.'''
        return self.spans.strip()



    def store(self):
        res = super().store()
        res.update({
            'wmode'     : self.wmode,
            'dir'       : self.dir,
            'line_break': self.line_break,
            'tab_stop'  : self.tab_stop,
            'spans'     : [
                span.store() for span in self.spans
            ]
        })

        return res


    def add(self, span_or_list):
        '''Add span list to current Line.
        
        Args:
            span_or_list (Span, Iterable): TextSpan or TextSpan list to add.
        '''
        if isinstance(span_or_list, Iterable):
            for span in span_or_list:
                self.add_span(span)
        else:
            self.add_span(span_or_list)


    def add_span(self, span:Element):
        '''Add span to current Line.'''
        self.spans.append(span)


    def intersects(self, rect):
        '''Create new Line object with spans contained in given bbox.
        
        Args:
            rect (fitz.Rect): Target bbox.
        
        Returns:
            Line: The created Line instance.
        '''
        # add line directly if fully contained in bbox
        if rect.contains(self.bbox):
            return self.copy()
        
        # new line with same text attributes
        line = Line({'wmode': self.wmode})
        line.dir = self.dir # update line direction relative to final CS

        # further check spans in line        
        for span in self.spans:
            contained_span = span.intersects(rect)
            line.add(contained_span)

        return line


    def make_docx(self, p):
        '''Create docx line, i.e. a run in ``python-docx``.'''
        # tab stop before this line to ensure horizontal position
        # Note it might need more than one tabs if multi-tabs are set for current paragraph
        if self.tab_stop: 
            for _ in range(self.tab_stop): p.add_run().add_tab()

        # create span -> run in paragraph
        for span in self.spans: span.make_docx(p)            

        # line break
        if self.line_break: p.add_run('\n')
            



# -*- coding: utf-8 -*-

'''Extract images from PDF.

Both raster images and vector graphics are considered:

* Normal images like jpeg or png could be extracted with method ``page.get_text('rawdict')`` 
  and ``Page.get_images()``. Note the process for png images with alpha channel.
* Vector graphics are actually composed of a group of paths, represented by operators like
  ``re``, ``m``, ``l`` and ``c``. They're detected by finding the contours with ``opencv``.
'''

import logging
import fitz
from ..common.Collection import Collection
from ..common.share import BlockType
from ..common.algorithm import (recursive_xy_cut, inner_contours, xy_project_profile)


class ImagesExtractor:
    def __init__(self, page:fitz.Page) -> None:
        '''Extract images from PDF page.
        
        Args:
            page (fitz.Page): pdf page to extract images.
        '''
        self._page = page
    

    def clip_page_to_pixmap(self, bbox:fitz.Rect=None, zoom:float=3.0):
        '''Clip page pixmap (without text) according to ``bbox``.

        Args:
            bbox (fitz.Rect, optional): Target area to clip. Defaults to None, i.e. entire page.
                Note that ``bbox`` depends on un-rotated page CS, while cliping page is based on
                the final page.
            zoom (float, optional): Improve resolution by this rate. Defaults to 3.0.

        Returns:
            fitz.Pixmap: The extracted pixmap.
        '''        
        # hide text 
        self._hide_page_text(self._page)
        
        if bbox is None:
            clip_bbox = self._page.rect
        
        # transform to the final bbox when page is rotated
        elif self._page.rotation:
            clip_bbox = bbox * self._page.rotation_matrix
            
        else:
            clip_bbox = bbox
        
        clip_bbox = clip_bbox & self._page.rect
        
        # improve resolution
        # - https://pymupdf.readthedocs.io/en/latest/faq.html#how-to-increase-image-resolution
        # - https://github.com/pymupdf/PyMuPDF/issues/181
        matrix = fitz.Matrix(zoom, zoom)

        return self._page.get_pixmap(clip=clip_bbox, matrix=matrix) # type: fitz.Pixmap


    def clip_page_to_dict(self, bbox:fitz.Rect=None, clip_image_res_ratio:float=3.0):
        '''Clip page pixmap (without text) according to ``bbox`` and convert to source image.

        Args:
            bbox (fitz.Rect, optional): Target area to clip. Defaults to None, i.e. entire page.
            clip_image_res_ratio (float, optional): Resolution ratio of clipped bitmap. Defaults to 3.0.

        Returns:
            list: A list of image raw dict.
        '''
        pix = self.clip_page_to_pixmap(bbox=bbox, zoom=clip_image_res_ratio)
        return self._to_raw_dict(pix, bbox)


    def extract_images(self, clip_image_res_ratio:float=3.0):
        '''Extract normal images with ``Page.get_images()``.

        Args:
            clip_image_res_ratio (float, optional): Resolution ratio of clipped bitmap. Defaults to 3.0.

        Returns:
            list: A list of extracted and recovered image raw dict.
        
        .. note::
            ``Page.get_images()`` contains each image only once, which may less than the real count of images in a page.
        '''
        # pdf document
        doc = self._page.parent
        rotation = self._page.rotation

        # The final view might be formed by several images with alpha channel only, as shown in issue-123. 
        # It's still inconvenient to extract the original alpha/mask image, as a compromise, extract the 
        # equivalent image by clipping the union page region for now.
        # https://github.com/dothinking/pdf2docx/issues/123

        # step 1: collect images: [(bbox, item), ..., ]
        ic = Collection()
        for item in self._page.get_images(full=True):
            # image item: (xref, smask, width, height, bpc, colorspace, ...)
            item = list(item)
            item[-1] = 0            
            
            # find all occurrences referenced to this image            
            rects = self._page.get_image_rects(item)
            unrotated_page_bbox = self._page.cropbox # note the difference to page.rect
            for bbox in rects:
                # ignore small images
                if bbox.get_area()<=4: continue

                # ignore images outside page
                if not unrotated_page_bbox.intersects(bbox): continue

                # collect images
                ic.append((bbox, item))

        # step 2: group by intersection
        fun = lambda a, b: a[0].intersects(b[0])
        groups = ic.group(fun)

        # step 3: check each group
        images = []
        for group in groups:
            # clip page with the union bbox of all intersected images
            if len(group) > 1:
                clip_bbox = fitz.Rect()
                for (bbox, item) in group: clip_bbox |= bbox
                raw_dict = self.clip_page_to_dict(clip_bbox, clip_image_res_ratio)
            
            else:
                bbox, item = group[0]
                # recover image
                pix = self._recover_pixmap(doc, item)

                # regarding images consist of alpha values only, i.e. colorspace is None,
                # the turquoise color shown in the PDF is not part of the image, but part of PDF background.
                # So, just to clip page pixmap according to the right bbox
                # https://github.com/pymupdf/PyMuPDF/issues/677
                alpha_only = not pix.colorspace
                if alpha_only:
                    raw_dict = self.clip_page_to_dict(bbox, clip_image_res_ratio)
                
                # rotate image with opencv if page is rotated
                else:
                    raw_dict = self._to_raw_dict(pix, bbox)
                    if rotation: 
                        raw_dict['image'] = self._rotate_image(pix, -rotation)

            images.append(raw_dict)

        return images    
        
    
    def detect_svg_contours(self, min_svg_gap_dx:float, min_svg_gap_dy:float, min_w:float, min_h:float):
        '''Find contour of potential vector graphics.

        Args:
            min_svg_gap_dx (float): Merge svg if the horizontal gap is less than this value.
            min_svg_gap_dy (float): Merge svg if the vertical gap is less than this value.
            min_w (float): Ignore contours if the bbox width is less than this value.
            min_h (float): Ignore contours if the bbox height is less than this value.

        Returns:
            list: A list of potential svg region: (external_bbox, inner_bboxes:list).
        '''
        import cv2 as cv

        # clip page and convert to opencv image
        pixmap = self.clip_page_to_pixmap(zoom=1.0)
        src = self._pixmap_to_cv_image(pixmap)

        # gray and binary
        gray = cv.cvtColor(src, cv.COLOR_BGR2GRAY)
        _, binary = cv.threshold(gray, 253, 255, cv.THRESH_BINARY_INV)
        
        # external bbox: split images with recursive xy cut
        external_bboxes = recursive_xy_cut(binary, min_dx=min_svg_gap_dx, min_dy=min_svg_gap_dy)        
        
        # inner contours
        grouped_inner_bboxes = [inner_contours(binary, bbox, min_w, min_h) for bbox in external_bboxes]

        # combined external and inner contours
        groups = list(zip(external_bboxes, grouped_inner_bboxes))
            
        
        # plot detected images for debug
        debug = False
        if debug:
            # plot projection profile for each sub-image
            for i, (x0, y0, x1, y1) in enumerate(external_bboxes):
                arr = xy_project_profile(src[y0:y1, x0:x1, :], binary[y0:y1, x0:x1])
                cv.imshow(f'sub-image-{i}', arr)

            for bbox, inner_bboxes in groups:
                # plot external bbox
                x0, y0, x1, y1 = bbox
                cv.rectangle(src, (x0, y0), (x1, y1), (255,0,0), 1)

                # plot inner bbox
                for u0, v0, u1, v1 in inner_bboxes:
                    cv.rectangle(src, (u0, v0), (u1, v1), (0,0,255), 1)

            cv.imshow("img", src)
            cv.waitKey(0)

        return groups


    @staticmethod
    def _to_raw_dict(image:fitz.Pixmap, bbox:fitz.Rect):
        '''Store Pixmap ``image`` to raw dict.

        Args:
            image (fitz.Pixmap): Pixmap to store.
            bbox (fitz.Rect): Boundary box the pixmap.

        Returns:
            dict: Raw dict of the pixmap.
        '''
        return {
            'type': BlockType.IMAGE.value,
            'bbox': tuple(bbox),
            'width': image.width,
            'height': image.height,
            'image': image.tobytes()
        }


    @staticmethod
    def _rotate_image(pixmap:fitz.Pixmap, rotation:int):
        '''Rotate image represented by image bytes.

        Args:
            pixmap (fitz.Pixmap): Image to rotate.
            rotation (int): Rotation angle.
        
        Return: image bytes.
        '''
        import cv2 as cv
        import numpy as np

        # convert to opencv image
        img = ImagesExtractor._pixmap_to_cv_image(pixmap)
        h, w = img.shape[:2] # get image height, width

        # calculate the center of the image
        x0, y0 = w//2, h//2

        # default scale value for now -> might be extracted from PDF page property    
        scale = 1.0

        # rotation matrix
        matrix = cv.getRotationMatrix2D((x0, y0), rotation, scale)

        # calculate the final dimension
        cos = np.abs(matrix[0, 0])
        sin = np.abs(matrix[0, 1])
    
        # compute the new bounding dimensions of the image
        W = int((h * sin) + (w * cos))
        H = int((h * cos) + (w * sin))
    
        # adjust the rotation matrix to take into account translation
        matrix[0, 2] += (W / 2) - x0
        matrix[1, 2] += (H / 2) - y0
        
        # perform the rotation holding at the center        
        rotated_img = cv.warpAffine(img, matrix, (W, H))

        # convert back to bytes
        _, im_png = cv.imencode('.png', rotated_img)
        return im_png.tobytes()


    @staticmethod
    def _hide_page_text(page:fitz.Page):
        '''Hide page text before clipping page.'''
        # NOTE: text might exist in both content stream and form object stream
        # - content stream, i.e. direct page content
        # - form object, i.e. contents referenced by this page
        xref_list = [xref for (xref, name, invoker, bbox) in page.get_xobjects()]
        xref_list.extend(page.get_contents())        

        # render Tr: set the text rendering mode
        # - 3: neither fill nor stroke the text -> invisible
        # read more:
        # - https://github.com/pymupdf/PyMuPDF/issues/257
        # - https://www.adobe.com/content/dam/acom/en/devnet/pdf/pdfs/pdf_reference_archives/PDFReference.pdf
        doc = page.parent # type: fitz.Document
        for xref in xref_list:
            stream = doc.xref_stream(xref).replace(b'BT', b'BT 3 Tr') \
                                             .replace(b'Tm', b'Tm 3 Tr') \
                                             .replace(b'Td', b'Td 3 Tr')
            doc.update_stream(xref, stream)
   
    @staticmethod
    def _recover_pixmap(doc:fitz.Document, item:list):
        """Restore pixmap with soft mask considered.
        
        References:

            * https://pymupdf.readthedocs.io/en/latest/document.html#Document.getPageImageList        
            * https://pymupdf.readthedocs.io/en/latest/faq.html#how-to-handle-stencil-masks
            * https://github.com/pymupdf/PyMuPDF/issues/670

        Args:
            doc (fitz.Document): pdf document.
            item (list): image instance of ``page.get_images()``.

        Returns:
            fitz.Pixmap: Recovered pixmap with soft mask considered.
        """
        # data structure of `item`:
        # (xref, smask, width, height, bpc, colorspace, ...)
        x = item[0]  # xref of PDF image
        s = item[1]  # xref of its /SMask

        # base image
        pix = fitz.Pixmap(doc, x)

        # reconstruct the alpha channel with the smask if exists
        if s > 0:
            mask = fitz.Pixmap(doc, s)
            if pix.alpha:
                temp = fitz.Pixmap(pix, 0)  # make temp pixmap w/o the alpha
                pix = None  # release storage
                pix = temp
            
            # check dimension
            if pix.width==mask.width and pix.height==mask.height:
                pix = fitz.Pixmap(pix, mask)  # now compose final pixmap
            else:
                logging.warning('Ignore image due to inconsistent size of color and mask pixmaps: %s', item)

        # we may need to adjust something for CMYK pixmaps here -> 
        # recreate pixmap in RGB color space if necessary
        # NOTE: pix.colorspace may be None for images with alpha channel values only
        if pix.colorspace and not pix.colorspace.name in (fitz.csGRAY.name, fitz.csRGB.name):
            pix = fitz.Pixmap(fitz.csRGB, pix)

        return pix


    @staticmethod
    def _pixmap_to_cv_image(pixmap:fitz.Pixmap):
        '''Convert fitz Pixmap to opencv image.

        Args:
            pixmap (fitz.Pixmap): PyMuPDF Pixmap.
        '''
        import cv2 as cv
        import numpy as np
        img_byte = pixmap.tobytes()
        return cv.imdecode(np.frombuffer(img_byte, np.uint8), cv.IMREAD_COLOR)


# -*- coding: utf-8 -*-

'''Object with a bounding box, e.g. Block, Line, Span.

Based on ``PyMuPDF``, the coordinates (e.g. bbox of ``page.get_text('rawdict')``) are generally 
provided relative to the un-rotated page; while this ``pdf2docx`` library works under real page 
coordinate system, i.e. with rotation considered. So, any instances created by this Class are 
always applied a rotation matrix automatically.

Therefore, the bbox parameter used to create ``Element`` instance MUST be relative to un-rotated
CS. If final coordinates are provided, should update it after creating an empty object::

    Element().update_bbox(final_bbox)

.. note::
    An exception is ``page.get_drawings()``, the coordinates are converted to real page CS already.
'''

import copy
import fitz
from .share import IText
from . import constants


class Element(IText):
    '''Boundary box with attribute in fitz.Rect type.'''

    # all coordinates are related to un-rotated page in PyMuPDF
    # e.g. Matrix(0.0, 1.0, -1.0, 0.0, 842.0, 0.0)
    ROTATION_MATRIX = fitz.Matrix(0.0) # rotation angle = 0 degree by default


    @classmethod
    def set_rotation_matrix(cls, rotation_matrix):
        """Set global rotation matrix.

        Args:
            Rotation_matrix (fitz.Matrix): target matrix
        """        
        if rotation_matrix and isinstance(rotation_matrix, fitz.Matrix):
            cls.ROTATION_MATRIX = rotation_matrix


    @classmethod
    def pure_rotation_matrix(cls):
        '''Pure rotation matrix used for calculating text direction after rotation.'''
        a,b,c,d,e,f = cls.ROTATION_MATRIX
        return fitz.Matrix(a,b,c,d,0,0)


    def __init__(self, raw:dict=None, parent=None):
        ''' Initialize Element and convert to the real (rotation considered) page coordinate system.'''        
        self.bbox = fitz.Rect()  # type: fitz.Rect
        self._parent = parent # type: Element

        # NOTE: Any coordinates provided in raw is in original page CS (without considering page rotation).
        if 'bbox' in (raw or {}):
            rect = fitz.Rect(raw['bbox']) * Element.ROTATION_MATRIX
            self.update_bbox(rect)


    def __bool__(self):
        '''Real object when bbox is defined.'''
        return bool(self.bbox)
    

    def __repr__(self): return f'{self.__class__.__name__}({tuple(self.bbox)})'


    # ------------------------------------------------
    # parent element
    # ------------------------------------------------
    @property
    def parent(self): return self._parent

    @parent.setter
    def parent(self, parent): self._parent = parent


    # ------------------------------------------------
    # bbox operations
    # ------------------------------------------------
    def copy(self):
        '''make a deep copy.'''
        # NOTE: can't serialize data because parent is an Object,
        # so set it None in advance.
        parent, self.parent = self._parent, None
        obj = copy.deepcopy(self)
        self._parent = parent # set back parent
        return obj


    def get_expand_bbox(self, dt:float):
        """Get expanded bbox with margin in both x- and y- direction.

        Args:
            dt (float): Expanding margin.

        Returns:
            fitz.Rect: Expanded bbox.
        
        .. note::
            This method creates a new bbox, rather than changing the bbox of itself.
        """        
        return self.bbox + (-dt, -dt, dt, dt)


    def update_bbox(self, rect):
        '''Update current bbox to specified ``rect``.
        
        Args:
            rect (fitz.Rect or list): bbox-like ``(x0, y0, x1, y1)`` in real page CS (with rotation considered).
        '''
        self.bbox = fitz.Rect([round(x,1) for x in rect])
        return self


    def union_bbox(self, e):
        """Update current bbox to the union with specified Element.

        Args:
            e (Element): The target to get union

        Returns:
            Element: self
        """ 
        return self.update_bbox(self.bbox | e.bbox)


    # --------------------------------------------
    # location relationship to other Element instance
    # -------------------------------------------- 
    def contains(self, e:'Element', threshold:float=1.0):
        """Whether given element is contained in this instance, with margin considered.

        Args:
            e (Element): Target element
            threshold (float, optional): Intersection rate. Defaults to 1.0. The larger, the stricter.

        Returns:
            bool: [description]
        """
        # NOTE the case bool(e)=True but e.bbox.get_area()=0
        S = e.bbox.get_area()
        if not S: return False 
        
        # it's not practical to set a general threshold to consider the margin, so two steps:
        # - set a coarse but acceptable area threshold,
        # - check the length in main direction strictly

        # A contains B => A & B = B
        intersection = self.bbox & e.bbox
        factor = round(intersection.get_area()/e.bbox.get_area(), 2)
        if factor<threshold: return False

        # check length
        if self.bbox.width >= self.bbox.height:
            return self.bbox.width+constants.MINOR_DIST >= e.bbox.width
        else:
            return self.bbox.height+constants.MINOR_DIST >= e.bbox.height
   

    def get_main_bbox(self, e, threshold:float=0.95):
        """If the intersection with ``e`` exceeds the threshold, return the union of these two elements; else return None.

        Args:
            e (Element): Target element.
            threshold (float, optional): Intersection rate. Defaults to 0.95.

        Returns:
            fitz.Rect: Union bbox or None.
        """
        bbox_1 = self.bbox
        bbox_2 = e.bbox if hasattr(e, 'bbox') else fitz.Rect(e)
        
        # areas
        b = bbox_1 & bbox_2
        if not b: return None # no intersection

        a1, a2, a = bbox_1.get_area(), bbox_2.get_area(), b.get_area()        

        # Note: if bbox_1 and bbox_2 intersects with only an edge, b is not empty but b.get_area()=0
        # so give a small value when they're intersected but the area is zero
        factor = a/min(a1,a2) if a else 1e-6
        return bbox_1 | bbox_2 if factor >= threshold else None


    def vertically_align_with(self, e, factor:float=0.0, text_direction:bool=True):
        '''Check whether two Element instances have enough intersection in vertical direction, i.e. perpendicular to reading direction.
        
        Args:
            e (Element): Object to check with
            factor (float, optional): Threshold of overlap ratio, the larger it is, the higher probability the two bbox-es are aligned.
            text_direction (bool, optional): Consider text direction or not. True by default, from left to right if False.

        Returns:
            bool: [description]
        
        Examples::

            +--------------+
            |              |
            +--------------+ 
                    L1
                    +-------------------+
                    |                   |
                    +-------------------+
                            L2
            
        An enough intersection is defined based on the minimum width of two boxes::
        
            L1+L2-L>factor*min(L1,L2)
        '''
        if not e or not bool(self): return False

        # text direction
        idx = 1 if text_direction and self.is_vertical_text else 0

        L1 = self.bbox[idx+2]-self.bbox[idx]
        L2 = e.bbox[idx+2]-e.bbox[idx]
        L = max(self.bbox[idx+2], e.bbox[idx+2]) - min(self.bbox[idx], e.bbox[idx])

        eps = 1e-3 # tolerant
        return L1+L2-L+eps >= factor*min(L1,L2)


    def horizontally_align_with(self, e, factor:float=0.0, text_direction:bool=True):
        '''Check whether two Element instances have enough intersection in horizontal direction, i.e. along the reading direction.
           
        Args:
            e (Element): Element to check with
            factor (float, optional): threshold of overlap ratio, the larger it is, the higher probability the two bbox-es are aligned.
            text_direction (bool, optional): consider text direction or not. True by default, from left to right if False.

        Examples::

            +--------------+
            |              | L1  +--------------------+
            +--------------+     |                    | L2
                                 +--------------------+
            
        An enough intersection is defined based on the minimum width of two boxes::
        
            L1+L2-L>factor*min(L1,L2)
        '''
        if not e or not bool(self): return False

        # text direction
        idx = 0 if text_direction and self.is_vertical_text else 1
        
        L1 = self.bbox[idx+2]-self.bbox[idx]
        L2 = e.bbox[idx+2]-e.bbox[idx]
        L = max(self.bbox[idx+2], e.bbox[idx+2]) - min(self.bbox[idx], e.bbox[idx])

        eps = 1e-3 # tolerant
        return L1+L2-L+eps >= factor*min(L1,L2)


    def in_same_row(self, e):
        """Check whether in same row/line with specified Element instance. With text direction considered.
           
           Taking horizontal text as an example:
           
           * yes: the bottom edge of each box is lower than the centerline of the other one;
           * otherwise, not in same row.

        Args:
            e (Element): Target object.

        Returns:
            bool: [description]
        
        .. note::
            The difference to method ``horizontally_align_with``: they may not in same line, though 
            aligned horizontally.
        """
        if not e or self.is_horizontal_text != e.is_horizontal_text:
            return False

        # normal reading direction by default
        idx = 1 if self.is_horizontal_text else 0

        c1 = (self.bbox[idx] + self.bbox[idx+2]) / 2.0
        c2 = (e.bbox[idx] + e.bbox[idx+2]) / 2.0
        res = c1<=e.bbox[idx+2] and c2<=self.bbox[idx+2] # Note y direction under PyMuPDF context
        return res


    # ------------------------------------------------
    # others
    # ------------------------------------------------
    def store(self):
        '''Store properties in raw dict.'''
        return { 'bbox': tuple([x for x in self.bbox]) }

    
    def plot(self, page, stroke:tuple=(0,0,0), width:float=0.5, fill:tuple=None, dashes:str=None):
        '''Plot bbox in PDF page for debug purpose.'''
        page.draw_rect(self.bbox, color=stroke, fill=fill, width=width, dashes=dashes, overlay=False, fill_opacity=0.5)


'''Base page with basic properties, e.g. width, height and margin.'''

class BasePage:
    def __init__(self, width:float=0.0, height:float=0.0, margin:tuple=None):
        '''Initialize page layout.

        Args:
            width (float, optional): Page width. Defaults to 0.0.
            height (float, optional): Page height. Defaults to 0.0.
            margin (tuple, optional): Page margin. Defaults to None.
        '''
        # page size and margin
        self.width = width
        self.height = height
        self.margin = margin or (0,) * 4
    

    @property
    def bbox(self): return (0.0, 0.0, self.width, self.height)


    @property
    def working_bbox(self):
        '''bbox with margin considered.'''
        x0, y0, x1, y1 = self.bbox
        L, R, T, B = self.margin
        return (x0+L, y0+T, x1-R, y1-B)


# -*- coding: utf-8 -*-

'''Row in a table.
'''

from docx.enum.table import WD_ROW_HEIGHT
from docx.shared import Pt
from .Cells import Cells
from ..common.Element import Element


class Row(Element):
    '''Row in a table.'''
    def __init__(self, raw:dict=None):
        if raw is None: raw = {}
        super().__init__(raw)

        # logical row height
        self.height = raw.get('height', 0.0)

        # cells in row
        self._cells = Cells(parent=self).restore(raw.get('cells', []))


    def __getitem__(self, idx):
        try:
            cell = self._cells[idx]
        except IndexError:
            msg = f'Cell index {idx} out of range'
            raise IndexError(msg)
        else:
            return cell

    def __iter__(self):
        return (cell for cell in self._cells)

    def __len__(self):
        return len(self._cells)


    def append(self, cell):
        '''Append cell to row and update bbox accordingly.'''
        self._cells.append(cell)


    def store(self):
        res = super().store()
        res.update({
            'height': self.height,
            'cells': self._cells.store()
        })

        return res


    def make_docx(self, table, idx_row:int):
        '''Create row of docx table.
        
        Args:
            table (Table): ``python-docx`` table instance.
            idx_row (int): Current row index.
        '''  
        # set row height
        docx_row = table.rows[idx_row]

        # to control the layout precisely, set `exact` value, rather than `at least` value
        # the associated steps in MS word: Table Properties -> Row -> Row height -> exactly
        docx_row.height_rule = WD_ROW_HEIGHT.EXACTLY
        
        # NOTE: row height is counted from center-line of top border to center line of bottom border
        docx_row.height = Pt(self.height)

        # set cell style and contents
        for idx_col in range(len(table.columns)):
            self._cells[idx_col].make_docx(table, (idx_row, idx_col))


# -*- coding: utf-8 -*-

'''Module to determine stream table borders.

Though no exact borders exist for stream table, it's better to simplify table structure by
aligning borders as more as possible. Taking vertical borders for example, it can be moved 
in a valid range in horizontal direction, but restricted by top and bottom borders in vertical 
direction. It's also true for horizontal borders.

Accordingly, introduce ``Border`` object, which has the following attributes:

* Valid range, e.g. ``(100, 250)``;
* Boundary borders, e.g. ``(top_border, bottom_border)`` for v-border,
  or ``(left_border, right_border)`` for h-border.

The target is to finalize the position in valid range, e.g. ``x=125`` for v-border with valid range 
``(120, 250)``. Then it's y-direction is determined by its boundary borders, where the y-coordinates 
are finalized in same logic. Finally, this border is fixed since both x- and y- directions are 
determined.

.. note::
    Consider horizontal and vertical borders only.
'''

from collections import defaultdict
from ..shape.Shapes import Shapes
from ..shape.Shape import Stroke
from ..common import constants
from ..common.share import RectType, rgb_value
from ..common.Collection import BaseCollection


class Border:
    '''Border for stream table.'''

    def __init__(self, border_type='hi', border_range:tuple=None, borders:tuple=None, reference:bool=False):
        '''Border for stream table.
        
        Args:
            border_type (str): border orientation/position, e.g. 'hi' represents horizontal inner border.
                * ``h/v``       - horizontal/vertical border;
                * ``t/b/l/r/i`` - outer border (top/bottom/left/right), or inner border
            border_range (tuple): Valid range, e.g. ``(x0, x1)`` for vertical border.
            borders (tuple): Boundary borders in ``Border`` type, e.g. 
                * top and bottom horizontal borders for current vertical border; 
                * left and right vertical borders for current horizontal border. 
            reference (bool): Reference border will not convert to real table border.
        '''
        # border type
        self.border_type = border_type.upper()

        # Whether the MAIN dimension (e.g. the vertical position for H-border) is determined.
        # NOTE: the other dimension (e.g. the left/right position for H-border) is determined 
        # by boundary borders, so a border is fully determined on condition that:
        # - it is finalized, 
        # - AND the boundary borders are also finalized. 
        self.finalized = False

        # Whether reference only border.
        self.is_reference = reference

        # valid range
        self.set_border_range(border_range)

        # boundary borders
        self.set_boundary_borders(borders)

        # the position to be finalized, e.g. y-coordinate for horizontal border
        self._value = None
        
        # border style
        self.width = constants.HIDDEN_W_BORDER
        self.color = 0 # black by default
        

    @property
    def is_horizontal(self): return 'H' in self.border_type

    @property
    def is_vertical(self): return 'V' in self.border_type

    @property
    def is_top(self): return 'T' in self.border_type

    @property
    def is_bottom(self): return 'B' in self.border_type

    @property
    def value(self):
        '''Finalized position, e.g. y-coordinate of horizontal border. 

        Average value if not finalized, but close to the table side for top and bottom 
        boundary borders.
        '''
        if self.finalized:
            return self._value
        else:
            avg = (self.LRange+self.URange)/2.0
            if self.is_top:
                return max(self.URange-3, avg)
            elif self.is_bottom:
                return min(self.LRange+3, avg)
            else:
                return avg

    @property
    def centerline(self):
        '''Center line of this border.'''
        if self.is_horizontal:
            return (self._LBorder.value, self.value, self._UBorder.value, self.value)
        else:
            return (self.value, self._LBorder.value, self.value, self._UBorder.value)


    def is_valid(self, value:float):
        '''Whether the given position locates in the valid border range.

            Args:
                value (float): Target position.
            
            Returns:
                bool: Valid position or not.
        '''
        # consider margin here, but pay attention to underline which may be counted
        return (self.LRange-constants.MINOR_DIST) <= value <= (self.URange+constants.MINOR_DIST) 


    def set_border_range(self, border_range:tuple=None):
        """Set border valid ranges.

        Args:
            border_range (tuple, optional): Lower/upper range to set. Defaults to None.

        Returns:
            Border: self
        """ 
        if border_range:
            x0, x1 = border_range
        else:
            x0, x1 = -9999, 9999
        self.LRange:float = x0
        self.URange:float = x1
        return self


    def set_boundary_borders(self, borders:tuple=None):
        """Set boundary borders.

        Args:
            borders (tuple, optional): Lower/upper boundary borders to set. Defaults to None.

        Returns:
            Border: self
        """
        if borders:
            lower_border, upper_border = borders
        else:
            lower_border, upper_border = None, None
        self._LBorder:Border = lower_border # left border, or top border
        self._UBorder:Border = upper_border # right border, or bottom border
        return self


    def get_boundary_borders(self):
        '''Get boundary borders.

        Returns:
            tuple: ``(lower b-border, upper b-border)``
        '''
        return (self._LBorder, self._UBorder)


    def finalize_by_value(self, value:float):
        '''Finalize border with given position.

        Args:
            value (float): Target position.
        '''
        # can be finalized only one time
        if self.finalized or not self.is_valid(value): return False

        self._value = value
        self.finalized = True
        self.is_reference = False

        return True


    def finalize_by_stroke(self, stroke:Stroke):
        '''Finalize border with specified stroke shape, which is generally a showing border-like shape.

        Args:
            stroke (Stroke): Target stroke to finalize this border.

        .. note::
            * The boundary borders may also be affected by this stroke shape.
            * The border-like stroke may be an underline or strike-through.      
        '''
        # NOTE: don't do this: `if self.finalized: continue`, 
        # because `self.finalized` just determined the main dimension, still need a chance to determine 
        # the other dimension.         

        if self.is_horizontal:
            # x0, x1, and y of h-stroke
            low_pos, upper_pos = stroke.x0, stroke.x1
            value = stroke.y0 # equal to stroke.y1
        else:
            # y0, y1 and x of v-stroke
            low_pos, upper_pos = stroke.y0, stroke.y1
            value = stroke.x0 # equal to stroke.x1

        # skip if not in valid range of a border
        if not self.is_valid(value): return

        # skip if not span in the border direction
        if low_pos > self._LBorder.URange and upper_pos < self._UBorder.LRange: return

        # now, finalize current border
        # NOTE: set border properties only if finalized by value successfully
        if self.finalize_by_value(value):
            self.color = stroke.color
            self.width = stroke.width
            stroke.type = RectType.BORDER  # update rect type as table border

        # and, give a chance to finalize boundary borders no matter this border is finalized or not,
        self._LBorder.finalize_by_value(low_pos)
        self._UBorder.finalize_by_value(upper_pos)


    def to_stroke(self):
        '''Convert to border stroke.'''
        # ignore if reference only
        if self.is_reference: return None

        stroke = Stroke({'color': self.color, 'width': self.width}).update_bbox(self.centerline)
        stroke.type = RectType.BORDER # set border style        
        return stroke


class Borders(BaseCollection):
    '''Collection of ``Border`` instances.'''

    def finalize(self, strokes:Shapes, fills:Shapes):
        '''Finalize the position of all borders.
        
        Args:
            strokes (Shapes): A group of explicit border strokes.
            fills (Shapes): A group of explicit cell shadings.

        .. note::
            A border is finalized in priority below:
            
            * Follow explicit stroke/border.
            * Follow explicit fill/shading.
            * Align h-borders or v-borders as more as possible to simplify the table structure.            
        '''
        # add dummy borders to be finalized by explicit strokes/fillings
        self._add_full_dummy_borders()

        # finalize borders by explicit strokes in first priority
        self._finalize_by_strokes(strokes)

        # finalize borders by explicit fillings in second priority
        tmp_strokes = []
        for fill in fills:
            # ignore determined filling or filling in white bg-color
            if fill.is_determined or fill.color == rgb_value((1,1,1)): continue

            x0, y0, x1, y1 = fill.bbox
            tmp_strokes.extend([
                Stroke().update_bbox((x0, y0, x1, y0)), # top
                Stroke().update_bbox((x0, y1, x1, y1)), # bottom
                Stroke().update_bbox((x0, y0, x0, y1)), # left
                Stroke().update_bbox((x1, y0, x1, y1))  # right
            ])
        self._finalize_by_strokes(tmp_strokes)

        # finalize borders by their layout finally (use an average position):
        # - un-finalized, and
        # - not reference-only borders        
        borders = list(filter(lambda border: not (border.finalized or border.is_reference), self._instances))

        # h-borders
        # NOTE: exclude the top and bottom boundary borders, since they'll be adjusted by
        # principle that minimizing the table region.
        h_borders = list(filter(
            lambda border: border.is_horizontal and 
                    not (border.is_top or border.is_bottom), borders))
        self._finalize_by_layout(h_borders)

        # v-borders
        v_borders = list(filter(lambda border: border.is_vertical, borders))
        self._finalize_by_layout(v_borders)

    
    def _finalize_by_strokes(self, strokes:list):
        '''Finalize borders by explicit strokes.'''
        for stroke in strokes:
            if stroke.is_determined: continue
            
            for border in self._instances:
                # horizontal stroke can finalize horizontal border only
                if stroke.horizontal != border.is_horizontal: continue
                
                border.finalize_by_stroke(stroke)                


    @staticmethod
    def _finalize_by_layout(borders:list):
        '''Finalize the position of all borders: 
        align borders as more as possible to simplify the table structure.

        Taking finalizing vertical borders for example:

        * initialize a list of x-coordinates, ``[x0, x1, x2, ...]``, with the interval points of each border
        * every two adjacent x-coordinates forms an interval for checking, ``[x0, x1]``, ``[x1, x2]``, ...
        * for each interval, count the intersection status of center point, ``x=(x0+x1)/2.0``, with all borders
        * sort center point with the count of intersections in decent order
        * finalize borders with x-coordinate of center points in sorting order consequently
        * terminate the process when all borders are finalized
        
        Args:
            borders (list): A list of ``Border`` instances.
        '''
        # collect interval points and sort in x-increasing order
        x_points = set()
        for border in borders:
            x_points.add(border.LRange)
            x_points.add(border.URange)
        
        x_points = list(x_points)
        x_points.sort()

        # check intersection status of each intervals
        x_status = [] # [(x, status), ...]
        for i in range(len(x_points)-1):
            x = (x_points[i]+x_points[i+1])/2.0 # center point
            s = [int(border.is_valid(x)) for border in borders]
            x_status.append((x,s))
            
        # sort per count since preferring passing through more borders
        x_status.sort(key=lambda item: sum(item[1]), reverse=True)

        # finalize borders
        num = len(borders)
        current_status = [0] * num
        for x, status in x_status:
            # terminate if all borders are finalized
            if sum(current_status) == num: break

            # only one line is allowed to pass through one border range -> sum(A.*B)=0
            #  e.g. A = [1,0,1,0], B=[0,1,0,0] -> valid
            #       A = [1,0,1,0], B =[1,0,0,0] -> invalid due to two lines passing through border 1
            duplicated = sum([c1*c2 for c1,c2 in zip(current_status, status)])
            if duplicated: continue

            # update current status
            current_status = [c1+c2 for c1,c2 in zip(current_status, status)]

            # now, finalize borders
            for border, border_status in zip(borders, status):
                if border_status: border.finalize_by_value(int(x))


    def _add_full_dummy_borders(self):
        '''Add reference borders to build full lattices.
        
        The original borders extracted from contents may be not able to represent the real 
        structure. Then, the reference borders has a chance to be finalized by explicit stroke 
        or fillings::

           +-------+---------+----------+
           +-------+---------+          +  <- shading in the first row, but not represented
           +-------+---------+          +  <- empty in these cells, so no borders extracted
           +-------+---------+----------+

        Add two dummy borders to form full lattices::

           +-------+---------+----------+
           +-------+---------+~~~~~~~~~~+  <- dummy borders
           +-------+---------+~~~~~~~~~~+
           +-------+---------+----------+
        '''
        h_borders = list(filter(lambda border: border.is_horizontal, self._instances))
        v_borders = list(filter(lambda border: border.is_vertical, self._instances))

        # group h-borders
        raw_borders_map = defaultdict(list)
        h_range_set = set()
        for border in h_borders:
            h_range = (border.LRange, border.URange)
            h_range_set.add(h_range)
            raw_borders_map[border.get_boundary_borders].append(h_range)

        # sort v-borders and try to add dummy h-borders between adjacent v-borders
        v_borders.sort(key=lambda border: border.value)
        for i in range(len(v_borders)-1):
            left, right = v_borders[i], v_borders[i+1]
            left_l_border, left_u_border = left.get_boundary_borders()
            right_l_border, right_u_border = right.get_boundary_borders()

            # the candidate v-borders must be connected
            if left_l_border!=right_l_border and left_u_border!=right_u_border: continue

            # get valid range
            lower_bound = max(left_l_border.LRange, right_l_border.LRange)
            upper_bound = min(left_u_border.URange, right_u_border.URange)

            # add dummy border if not exist
            raw_borders = raw_borders_map.get((left,right), []) # existed borders
            for h_range in h_range_set:
                # ignore if existed
                if h_range in raw_borders: continue

                # dummy border must in valid range
                if h_range[0]>upper_bound or h_range[1]<lower_bound: continue

                h_border = Border('HI', h_range, (left, right), reference=True)
                self._instances.append(h_border)



# -*- coding: utf-8 -*-

'''
A wrapper of PyMuPDF Page as page engine.
'''

import logging
from .RawPage import RawPage
from ..image.ImagesExtractor import ImagesExtractor
from ..shape.Paths import Paths
from ..common.constants import FACTOR_A_HALF
from ..common.Element import Element
from ..common.share import (RectType, debug_plot)
from ..common.algorithm import get_area


class RawPageFitz(RawPage):
    '''A wrapper of ``fitz.Page`` to extract source contents.'''

    def extract_raw_dict(self, **settings):
        raw_dict = {}
        if not self.page_engine: return raw_dict

        # actual page size
        *_, w, h = self.page_engine.rect # always reflecting page rotation
        raw_dict.update({ 'width' : w, 'height': h })
        self.width, self.height = w, h

        # pre-processing layout elements. e.g. text, images and shapes
        text_blocks = self._preprocess_text(**settings)
        raw_dict['blocks'] = text_blocks

        image_blocks = self._preprocess_images(**settings)
        raw_dict['blocks'].extend(image_blocks)
        
        shapes, images =  self._preprocess_shapes(**settings)
        raw_dict['shapes'] = shapes
        raw_dict['blocks'].extend(images)

        hyperlinks = self._preprocess_hyperlinks()
        raw_dict['shapes'].extend(hyperlinks)        
       
        # Element is a base class processing coordinates, so set rotation matrix globally
        Element.set_rotation_matrix(self.page_engine.rotation_matrix)

        return raw_dict
    

    def _preprocess_text(self, **settings):
        '''Extract page text and identify hidden text. 
        
        NOTE: All the coordinates are relative to un-rotated page.

            https://pymupdf.readthedocs.io/en/latest/page.html#modifying-pages
            https://pymupdf.readthedocs.io/en/latest/functions.html#Page.get_texttrace
            https://pymupdf.readthedocs.io/en/latest/textpage.html
        '''
        ocr = settings['ocr']
        if ocr==1: raise SystemExit("OCR feature is planned but not implemented yet.")

        # all text blocks no matter hidden or not
        raw = self.page_engine.get_text('rawdict', flags=64)
        text_blocks = raw.get('blocks', [])

        # potential UnicodeDecodeError issue when trying to filter hidden text:
        # https://github.com/dothinking/pdf2docx/issues/144
        # https://github.com/dothinking/pdf2docx/issues/155
        try:
            spans = self.page_engine.get_texttrace()
        except SystemError:
            logging.warning('Ignore hidden text checking due to UnicodeDecodeError in upstream library.')
            spans = []
        
        if not spans: return text_blocks

        # ignore hidden text if ocr=0, while extract only hidden text if ocr=2
        if ocr==2:
            f = lambda span: span['type']!=3  # find displayed text and ignore it
        else:
            f = lambda span: span['type']==3  # find hidden text and ignore it
        filtered_spans = list(filter(f, spans))
        
        def span_area(bbox):
            x0, y0, x1, y1 = bbox
            return (x1-x0) * (y1-y0)

        # filter blocks by checking span intersection: mark the entire block if 
        # any span is matched
        blocks = []
        for block in text_blocks:
            intersected = False
            for line in block['lines']:
                for span in line['spans']:
                    for filter_span in filtered_spans:
                        intersected_area = get_area(span['bbox'], filter_span['bbox'])
                        if intersected_area / span_area(span['bbox']) >= FACTOR_A_HALF \
                            and span['font']==filter_span['font']:
                            intersected = True
                            break
                    if intersected: break # skip further span check if found
                if intersected: break     # skip further line check

            # keep block if no any intersection with filtered span
            if not intersected: blocks.append(block)

        return blocks


    def _preprocess_images(self, **settings):
        '''Extract image blocks. Image block extracted by ``page.get_text('rawdict')`` doesn't 
        contain alpha channel data, so it has to get page images by ``page.get_images()`` and 
        then recover them. Note that ``Page.get_images()`` contains each image only once, i.e., 
        ignore duplicated occurrences.
        '''
        # ignore image if ocr-ed pdf: get ocr-ed text only
        if settings['ocr']==2: return []
        
        return ImagesExtractor(self.page_engine).extract_images(settings['clip_image_res_ratio'])


    def _preprocess_shapes(self, **settings):
        '''Identify iso-oriented paths and convert vector graphic paths to pixmap.'''
        paths = self._init_paths(**settings)
        return paths.to_shapes_and_images(
            settings['min_svg_gap_dx'], 
            settings['min_svg_gap_dy'], 
            settings['min_svg_w'], 
            settings['min_svg_h'], 
            settings['clip_image_res_ratio'])
    

    @debug_plot('Source Paths')
    def _init_paths(self, **settings):
        '''Initialize Paths based on drawings extracted with PyMuPDF.'''
        raw_paths = self.page_engine.get_cdrawings()
        return Paths(parent=self).restore(raw_paths)
    

    def _preprocess_hyperlinks(self):
        """Get source hyperlink dicts.

        Returns:
            list: A list of source hyperlink dict.
        """
        hyperlinks = []
        for link in self.page_engine.get_links():
            if link['kind']!=2: continue # consider internet address only
            hyperlinks.append({
                'type': RectType.HYPERLINK.value,
                'bbox': tuple(link['from']),
                'uri' : link['uri']
            })

        return hyperlinks



'''
Load :py:class:`~pdf2docx.page.RawPage` with specified pdf engine, 
e.g. PyMuPDF, pdfminer.six. For now, only PyMuPDF is implemented.
'''

from .RawPageFitz import RawPageFitz


class RawPageFactory:

    MAP = {
        'PYMUPDF': RawPageFitz
    }

    @classmethod
    def create(cls, page_engine, backend:str='pymupdf'):
        '''Create RawPage class with specified backend.'''
        klass = cls.MAP.get(backend.upper(), None)
        if not klass:
            raise TypeError(f'Page with pdf engine "{backend}" is not implemented yet.')
        else:
            return klass(page_engine=page_engine)
        


# -*- coding: utf-8 -*-

'''A group of Row objects in a table.
'''

from .Row import Row
from ..common.Collection import ElementCollection


class Rows(ElementCollection):
    '''A group of Rows.'''

    def restore(self, raws:list):
        """Restore Rows from source dicts.

        Args:
            raws (list): A list of source dicts representing each row.

        Returns:
            Rows: self
        """        
        for raw in raws:
            row = Row(raw)
            self.append(row)
        return self



# -*- coding: utf-8 -*-

'''Column of Section.

In most cases, one section per page. But in case multi-columns page, sections are used 
to distinguish these different layouts.

.. note::
    Currently, support at most two columns.

::

    {
        'bbox': (x0, y0, x1, y1),
        'blocks': [{
            ... # block instances
        }, ...],
        'shapes': [{
            ... # shape instances
        }, ...]
    }
'''

from ..common.Collection import Collection
from ..common.Element import Element
from ..layout.Layout import Layout
from ..shape.Shape import Shape
from ..text.Line import Line


class Column(Element, Layout):

    def __init__(self, blocks=None, shapes=None):
        '''Initialize empty column.'''
        # Call the first parent class constructor only if omitting constructor. 
        # Unified constructor should be used (with *args, **kwargs) if using super().__init__().
        Element.__init__(self)
        Layout.__init__(self, blocks, shapes)


    @property
    def working_bbox(self): return self.bbox


    def add_elements(self, elements:Collection):
        '''Add candidate elements, i.e. lines or shapes, to current column.'''
        blocks = [e for e in elements if isinstance(e, Line)]
        shapes = [e for e in elements if isinstance(e, Shape)]
        self.assign_blocks(blocks)
        self.assign_shapes(shapes)


    def store(self):
        '''Store parsed section layout in dict format.'''
        res = Element.store(self)
        res.update(Layout.store(self))
        return res


    def restore(self, raw:dict):
        '''Restore Column from raw dict.'''
        self.update_bbox(raw.get('bbox', (0,)*4))
        super().restore(raw)
        return self


    def make_docx(self, doc):
        '''Create Section Column in docx. 

        Args:
            doc (Document): ``python-docx`` document object
        '''
        self.blocks.make_docx(doc)






from collections import deque
import numpy as np
import cv2 as cv


# -------------------------------------------------------------------------------------------
# Intersection area of two iso-oriented rectangles
# -------------------------------------------------------------------------------------------
def get_area(bbox_1:tuple, bbox_2:tuple):
    x0, y0, x1, y1 = bbox_1
    u0, v0, u1, v1 = bbox_2

    # width of intersected area
    w = (x1-x0) + (u1-u0) - (max(x1, u1)-min(x0, u0))
    if w<=0: return 0

    # height of intersected area
    h = (y1-y0) + (v1-v0) - (max(y1, v1)-min(y0, v0))
    if h<=0: return 0

    return w*h


# -------------------------------------------------------------------------------------------
# Breadth First Search method for graph
# -------------------------------------------------------------------------------------------
def graph_bfs(graph): 
    '''Breadth First Search graph (may be disconnected graph).
    
    Args:
        graph (list): GRAPH represented by adjacent list, [set(1,2,3), set(...), ...]
    
    Returns:
        list: A list of connected components
    '''
    # search graph
    # NOTE: generally a disconnected graph
    counted_indexes = set() # type: set[int]
    groups = []
    for i in range(len(graph)):
        if i in counted_indexes: continue
        # connected component starts...
        indexes = set(_graph_bfs_from_node(graph, i))
        groups.append(indexes)
        counted_indexes.update(indexes)

    return groups


def _graph_bfs_from_node(graph, start):
    '''Breadth First Search connected graph with start node.
    
    Args:
        graph (list): GRAPH represented by adjacent list, [set(1,2,3), set(...), ...].
        start (int): Index of any start vertex.
    '''
    search_queue = deque()    
    searched = set()

    search_queue.append(start)
    while search_queue:
        cur_node = search_queue.popleft()
        if cur_node in searched: continue
        yield cur_node
        searched.add(cur_node)
        for node in graph[cur_node]:
            search_queue.append(node)



# -------------------------------------------------------------------------------------------
# Implementation of solving Rectangle-Intersection Problem according to algorithm proposed in
# paper titled "A Rectangle-Intersection Algorithm with Limited Resource Requirements".
# https://ieeexplore.ieee.org/document/5578313
# 
# - Input
# The rectangle is represented by its corner points, (x0, y0, x1, y1)
# 
# - Output
# The output is an Adjacent List of each rect, which could be used to initialize a GRAPH.
# -------------------------------------------------------------------------------------------
# procedure report(S, n)
# 1 Let V be the list of x-coordinates of the 2n vertical edges in S sorted in non-decreasing order.
# 2 Let H be the list of n y-intervals corresponding to the bottom and top y-coordinates of each rectangle.
# 3 Sort the elements of H in non-decreasing order by their bottom y-coordinates.
# 4 Call procedure detect(V, H, 2n).

def solve_rects_intersection(V:list, num:int, index_groups:list):
    '''Implementation of solving Rectangle-Intersection Problem.

    Performance::

        O(nlog n + k) time and O(n) space, where k is the count of intersection pairs.

    Args:
        V (list): Rectangle-related x-edges data, [(index, Rect, x), (...), ...].
        num (int): Count of V instances, equal to len(V).
        index_groups (list): Target adjacent list for connectivity between rects.
    
    Procedure ``detect(V, H, m)``::
    
        if m < 2 then return else
        - let V1 be the first ⌊m/2⌋ and let V2 be the rest of the vertical edges in V in the sorted order;
        - let S11 and S22 be the set of rectangles represented only in V1 and V2 but not spanning V2 and V1, respectively;
        - let S12 be the set of rectangles represented only in V1 and spanning V2; 
        - let S21 be the set of rectangles represented only in V2 and spanning V1
        - let H1 and H2 be the list of y-intervals corresponding to the elements of V1 and V2 respectively
        - stab(S12, S22); stab(S21, S11); stab(S12, S21)
        - detect(V1, H1, ⌊m/2⌋); detect(V2, H2, m − ⌊m/2⌋)
    '''
    if num < 2: return
    
    # start/end points of left/right intervals
    center_pos = int(num/2.0)
    X0, X, X1 = V[0][-1], V[center_pos-1][-1], V[-1][-1] 

    # split into two groups
    left = V[0:center_pos]
    right = V[center_pos:]

    # filter rects according to their position to each intervals
    S11 = list(filter( lambda item: item[1][2]<=X, left ))
    S12 = list(filter( lambda item: item[1][2]>=X1, left ))
    S22 = list(filter( lambda item: item[1][0]>X, right ))
    S21 = list(filter( lambda item: item[1][0]<=X0, right ))
    
    # intersection in x-direction is fulfilled, so check y-direction further
    _stab(S12, S22, index_groups)
    _stab(S21, S11, index_groups)
    _stab(S12, S21, index_groups)

    # recursive process
    solve_rects_intersection(left,  center_pos,     index_groups)
    solve_rects_intersection(right, num-center_pos, index_groups)


def _stab(S1:list, S2:list, index_groups:list):
    '''Check interval intersection in y-direction.
    
    Procedure ``stab(A, B)``::
        i := 1; j := 1
        while i ≤ |A| and j ≤ |B|
            if ai.y0 < bj.y0 then
            k := j
            while k ≤ |B| and bk.y0 < ai.y1
                reportPair(air, bks)
                k := k + 1
            i := i + 1
            else
            k := i
            while k ≤ |A| and ak.y0 < bj.y1
                reportPair(bjs, akr)
                k := k + 1
            j := j + 1
    '''
    if not S1 or not S2: return

    # sort
    S1.sort(key=lambda item: item[1][1])
    S2.sort(key=lambda item: item[1][1])

    i, j = 0, 0
    while i<len(S1) and j<len(S2):
        m, a, _ = S1[i]
        n, b, _ = S2[j]
        if a[1] < b[1]:
            k = j
            while k<len(S2) and S2[k][1][1] < a[3]:
                _report_pair(int(m/2), int(S2[k][0]/2), index_groups)
                k += 1
            i += 1
        else:
            k = i
            while k<len(S1) and S1[k][1][1] < b[3]:
                _report_pair(int(S1[k][0]/2), int(n/2), index_groups)
                k += 1
            j += 1


def _report_pair(i:int, j:int, index_groups:list):
    '''add pair (i,j) to adjacent list.'''
    index_groups[i].add(j)
    index_groups[j].add(i)



# -------------------------------------------------------------------------------------------
# Implementation of recursive X-Y cut algorithm, which is:
# a top-down page segmentation technique that decomposes a document image recursively into a 
# set of rectangular blocks.
# 
# - https://en.wikipedia.org/wiki/Recursive_X-Y_cut
# - Recursive X-Y Cut using Bounding Boxes of Connected Components by Jaekyu Ha, 
#   Robert M.Haralick and Ihsin T. Phillips
# -------------------------------------------------------------------------------------------
def recursive_xy_cut(img_binary:np.array, 
                    min_w:float=0.0, min_h:float=0.0, 
                    min_dx:float=15.0, min_dy:float=15.0): 
    '''Split image with recursive xy-cut algorithm.
    
    Args:
        img_binary (np.array): Binarized image with interesting region (255) and empty region (0).
        min_w (float): Ignore bbox if the width is less than this value.
        min_h (float): Ignore bbox if the height is less than this value.
        min_dx (float): Merge two bbox-es if the x-gap is less than this value.
        min_dy (float): Merge two bbox-es if the y-gap is less than this value.
    
    Returns:
        list: bbox (x0, y0, x1, y1) of split blocks.
    '''
    def xy_cut(arr:np.array, top_left:tuple, res:list, 
                    min_w:float, min_h:float, min_dx:float, min_dy:float):
        x0, y0 = top_left
        h, w = arr.shape
        # cut along x-direction
        projection = np.count_nonzero(arr==255, axis=1)
        pos_y = _split_projection_profile(projection, min_w, min_dy)
        if not pos_y: return        

        # cut along y-direction for each part
        arr_y0, arr_y1 = pos_y
        for r0, r1 in zip(arr_y0, arr_y1):
            x_arr = arr[r0:r1, 0:w]
            projection = np.count_nonzero(x_arr==255, axis=0)
            pos_x = _split_projection_profile(projection, min_h, min_dx)
            if not pos_x: continue
            
            # determined the block bbox
            arr_x0, arr_x1 = pos_x
            if len(arr_x0)==1:
                res.append((x0+arr_x0[0], y0+r0, x0+arr_x1[0], y0+r1))
                continue
            
            # xy-cut recursively if the count of blocks > 1
            for c0, c1 in zip(arr_x0, arr_x1):
                y_arr = arr[r0:r1, c0:c1]
                top_left = (x0+c0, y0+r0)
                xy_cut(y_arr, top_left, res, min_w, min_h, min_dx, min_dy)

    # do xy-cut recursively
    res = []
    xy_cut(arr=img_binary, top_left=(0, 0), res=res, 
            min_w=min_w, min_h=min_h, min_dx=min_dx, min_dy=min_dy)
    return res


def _split_projection_profile(arr_values:np.array, min_value:float, min_gap:float):
    '''Split projection profile:

    ```
                              ┌──┐
         arr_values           │  │       ┌─┐───
             ┌──┐             │  │       │ │ |
             │  │             │  │ ┌───┐ │ │min_value
             │  │<- min_gap ->│  │ │   │ │ │ |
         ────┴──┴─────────────┴──┴─┴───┴─┴─┴─┴───
         0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16
    ```

    Args:
        arr_values (np.array): 1-d array representing the projection profile.
        min_value (float): Ignore the profile if `arr_value` is less than `min_value`.
        min_gap (float): Ignore the gap if less than this value.

    Returns:
        tuple: Start indexes and end indexes of split groups.
    '''
    # all indexes with projection height exceeding the threshold
    arr_index = np.where(arr_values>min_value)[0]
    if not len(arr_index): return

    # find zero intervals between adjacent projections
    # |  |                    ||
    # ||||<- zero-interval -> |||||
    arr_diff = arr_index[1:] - arr_index[0:-1]
    arr_diff_index = np.where(arr_diff>min_gap)[0]
    arr_zero_intvl_start = arr_index[arr_diff_index]
    arr_zero_intvl_end = arr_index[arr_diff_index+1]

    # convert to index of projection range:
    # the start index of zero interval is the end index of projection
    arr_start = np.insert(arr_zero_intvl_end, 0, arr_index[0])
    arr_end = np.append(arr_zero_intvl_start, arr_index[-1])
    arr_end += 1 # end index will be excluded as index slice

    return arr_start, arr_end


def inner_contours(img_binary:np.array, bbox:tuple, min_w:float, min_h:float):
    '''Inner contours of current region, especially level 2 contours of the default opencv tree hirerachy.

    Args:
        img_binary (np.array): Binarized image with interesting region (255) and empty region (0).
        bbox (tuple): The external bbox.
        min_w (float): Ignore contours if the bbox width is less than this value.
        min_h (float): Ignore contours if the bbox height is less than this value.

    Returns:
        list: A list of bbox-es of inner contours.
    '''
    
    # find both external and inner contours of current region
    x0, y0, x1, y1 = bbox
    arr = np.zeros(img_binary.shape, dtype=np.uint8)
    arr[y0:y1, x0:x1] = img_binary[y0:y1, x0:x1]
    contours, hierarchy = cv.findContours(arr, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE)

    # check first three level contours:    
    # * level-0, i.e. table bbox
    # * level-1, i.e. cell bbox
    # * level-2, i.e. region within cell
    # NOTE: only one dimension, i.e. the second, to be decided, so the
    # return value of np.where is a len==1 tuple
    level_0 = np.where(hierarchy[0,:,3]==-1)[0]    
    level_1 = np.where(np.isin(hierarchy[0,:,3], level_0))[0]    
    level_2 = np.where(np.isin(hierarchy[0,:,3], level_1))[0]

    # In general, we focus on only level 2, but considering edge case: level 2 contours 
    # might be counted as level 1 incorrectly, e.g. test/samples/demo-table-close-underline.pdf. 
    # So, get first the concerned level 1 contours, i.e. those contained by other level 1 contour.
    def contains(bbox1, bbox2):
        x0, y0, x1, y1 = bbox1
        u0, v0, u1, v1 = bbox2
        return u0>=x0 and v0>=y0 and u1<=x1 and v1<=y1

    level_1_bbox_list, res_level_1, res = [], [], []
    for i in level_1:
        x, y, w, h = cv.boundingRect(contours[i])
        if w<min_w or h<min_h: continue
        level_1_bbox_list.append((x, y, x+w, y+h))

    for bbox1 in level_1_bbox_list:
        for bbox2 in level_1_bbox_list:
            if bbox1==bbox2: continue # skip itself
            if contains(bbox1, bbox2):
                res_level_1.append(bbox2)
                res.append(bbox2)

    # now level 2: with contours contained in `res_level_1` excluded
    def contained_in_concerned_level_1(bbox):
        for level_1_bbox in res_level_1:
            if contains(level_1_bbox, bbox): return True
        return False

    for i in level_2:
        x, y, w, h = cv.boundingRect(contours[i])
        level_2_bbox = (x, y, x+w, y+h)
        if w<min_w or h<min_h: continue
        if contained_in_concerned_level_1(level_2_bbox): continue
        res.append(level_2_bbox)
    
    return res


def xy_project_profile(img_source:np.array, img_binary:np.array, gap:int=5, dw:int=None, dh:int=None):   
    '''Projection profile along x and y direction.

    ```
           ┌────────────────┐
        dh │                │
           └────────────────┘
                 gap
           ┌────────────────┐ ┌───┐
           │                │ │   │
         h │     image      │ │   │
           │                │ │   │
           └────────────────┘ └───┘
                    w           dw
    ```

    Args:
        img_source (np.array): Source image, e.g. RGB mode.
        img_binary (np.array): Binarized image.
        gap (int, optional): Gap between sub-graph. Defaults to 5.
        dw (int, optional): Graph height of x projection profile. Defaults to None.
        dh (int, optional): Graph height of y projection profile. Defaults to None.

    Returns:
        np.array: The combined graph data.
    '''
    # combined image
    h, w = img_binary.shape
    dh = dh or max(int(h/3), 15)
    dw = dw or max(int(w/3), 15)
    arr = 255*np.ones((h+dh+gap, w+dw+gap, 3), dtype=np.uint8)

    # source image
    arr[dh+gap:dh+gap+h, 0:w, :] = img_source

    # x projection
    vals = np.count_nonzero(img_binary==255, axis=1)
    for i, val in enumerate(vals):
        c = int(val/w*dw)
        arr[i+dh+gap, w+gap:w+gap+int(c), :] = 0
    
    # y projection
    vals = np.count_nonzero(img_binary==255, axis=0)
    for i, val in enumerate(vals):
        r = int(val/h*dh)
        arr[dh-r:dh, i, :] = 0

    return arr




# -*- coding: utf-8 -*-

'''Collection of Cell instances.
'''

from .Cell import Cell
from ..common.Collection import ElementCollection


class Cells(ElementCollection):
    '''A group of Cells.'''
    def restore(self, raws:list):
        '''Restore Cells from source dict.

        Args:
            raws (list): A list of source dict.
        '''
        for raw in raws:
            cell = Cell(raw)
            self.append(cell)
        return self
    
    def append(self, cell:Cell):
        '''Override. Append a cell (allow empty cell, i.e. merged cells) and update bbox accordingly.'''
        self._instances.append(cell)
        self._update_bbox(cell)
        cell.parent = self._parent # set parent



'''
A group of text elements, distinguished to ``Shape`` elements. For instance, ``TextBlock``, 
``ImageBlock`` or ``TableBlock`` after parsing, while ``Line`` instances at the beginning, 
and a combination of ``Line`` and ``TableBlock`` during parsing process.
'''

import logging
from docx.shared import Pt
from ..common import constants
from ..common.Collection import ElementCollection
from ..common.share import (BlockType, lower_round, rgb_value)
from ..common.Block import Block
from ..common.docx import (reset_paragraph_format, delete_paragraph)
from ..text.TextBlock import TextBlock
from ..text.TextSpan import TextSpan
from ..text.Line import Line
from ..text.Lines import Lines
from ..table.Cell import Cell
from ..image.ImageBlock import ImageBlock
from ..table.TableBlock import TableBlock


class Blocks(ElementCollection):
    '''Block collections.'''
    def __init__(self, instances:list=None, parent=None):
        ''' A collection of text based elements, e.g. lines, images or blocks.'''
        super().__init__(instances, parent)
        self._floating_image_blocks = []


    def _update_bbox(self, block:Block):
        '''Override. The parent is ``Layout``, which is not necessary to update its bbox. 
        So, do nothing but required here.
        '''
        pass


    @property
    def floating_image_blocks(self):
        return self._floating_image_blocks


    @property
    def lattice_table_blocks(self):
        '''Get lattice table blocks contained in this Collection.'''
        return list(filter(
            lambda block: isinstance(block, Block) and block.is_lattice_table_block, self._instances))


    @property
    def stream_table_blocks(self):
        '''Get stream table blocks contained in this Collection.'''
        return list(filter(
            lambda block: isinstance(block, Block) and block.is_stream_table_block, self._instances))


    @property
    def table_blocks(self):
        '''Get table blocks contained in this Collection.'''
        return list(filter(
            lambda block: isinstance(block, Block) and block.is_table_block, self._instances))


    @property
    def inline_image_blocks(self):
        '''Get inline image blocks contained in this Collection.'''
        return list(filter(
            lambda block: isinstance(block, Block) and block.is_inline_image_block, self._instances))


    @property
    def text_blocks(self):
        '''Get text/image blocks contained in this Collection.'''
        return list(filter(
            lambda block: isinstance(block, Block) and block.is_text_image_block, self._instances))


    def restore(self, raws:list):
        '''Clean current instances and restore them from source dict.
        ImageBlock is converted to ImageSpan contained in TextBlock.

        Args:
            raws (list): A list of raw dicts representing text/image/table blocks.

        Returns:
            Blocks: self
        '''
        self.reset()  # clean current instances
        for raw_block in raws:
            block_type = raw_block.get('type', -1) # type: int
            
            # inline image block -> text block
            if block_type == BlockType.IMAGE.value:
                block = ImageBlock(raw_block).to_text_block()
            
            # text block
            elif block_type == BlockType.TEXT.value:
                block = TextBlock(raw_block)
            
            # table block
            elif block_type in (BlockType.LATTICE_TABLE.value, BlockType.STREAM_TABLE.value):
                block = TableBlock(raw_block)
            
            else:
                block = None            
            
            # add to list
            self.append(block)
        
        return self


    def clean_up(self, float_image_ignorable_gap:float, line_overlap_threshold:float):
        '''Clean up blocks in page level.

        * convert to lines
        * remove lines out of page
        * remove transformed text: text direction is not (1, 0) or (0, -1)
        * remove empty lines

        Args:
            float_image_ignorable_gap (float): Regarded as float image if the intersection exceeds this value.
            line_overlap_threshold (float): remove line if the intersection exceeds this value.

        .. note::
            The block structure extracted from ``PyMuPDF`` might be unreasonable, e.g. 
            * one real paragraph is split into multiple blocks; or
            * one block consists of multiple real paragraphs
        '''
        if not self._instances: return

        # convert to lines
        instances = []
        for block in self._instances:
            if not isinstance(block, (ImageBlock, TextBlock)): continue
            instances.extend(block.lines)
        
        # delete invalid lines
        page_bbox = self.parent.working_bbox
        f = lambda line: line.bbox.intersects(page_bbox) and \
                        not line.white_space_only and (
                            line.is_horizontal_text or line.is_vertical_text)
        instances = list(filter(f, instances))

        # delete redundant blanks
        for line in instances: line.strip()

        # detect floating images
        self.reset(instances) \
            ._identify_floating_images(float_image_ignorable_gap) \
            ._remove_overlapped_lines(line_overlap_threshold)


    def assign_to_tables(self, tables:list):
        '''Add blocks (line or sub-table) to associated cells of given tables.

        Args:
            tables (list): A list of TableBlock instances.
        '''        
        if not tables: return

        # assign blocks to table region        
        blocks_in_tables = [[] for _ in tables] # type: list[list[Line|TableBlock]]
        blocks = []   # type: list[Line|TableBlock]
        for block in self._instances:
            self._assign_block_to_tables(block, tables, blocks_in_tables, blocks)

        # assign blocks to associated cells
        for table, blocks_in_table in zip(tables, blocks_in_tables):
            # no contents for this table
            if not blocks_in_table: continue
            table.assign_blocks(blocks_in_table)

        # sort in natural reading order and update layout blocks
        blocks.extend(tables)
        self.reset(blocks)


    def collect_stream_lines(self, potential_shadings:list, line_separate_threshold:float):
        '''Collect elements in Line level (line or table bbox), which may contained in a stream table region.
        
        Table may exist on the following conditions:

        * blocks in a row don't follow flow layout; or
        * block is contained in potential shading
        
        Args:
            potential_shadings (list): a group of shapes representing potential cell shading
            line_separate_threshold (float): two separate lines if the x-distance exceeds this value
        
        Returns:
            list: A list of Lines. Each group of Lines represents a potential table.
        
        .. note::
            ``PyMuPDF`` may group multi-lines in a row as a text block while each line belongs to different
            cell. So, it's required to deep into line level.
        '''
        # group lines by row: any intersection would be counted as same group
        # NOTE: consider real text direction when all lines have same orientation, i.e. either horizontal or 
        # vertical; otherwise, consider the default direction, i.e. horizontal.
        rows = self.group_by_rows(text_direction=not self.is_mix_text)

        # get sub-lines from block: line or table block
        def sub_line(block):
            return block if isinstance(block, Line) else Line().update_bbox(block.outer_bbox)
        
        # exclude potential shading in white bg-color
        shadings_exclude_white = list(filter(
            lambda shape: not shape.is_determined and shape.color!=rgb_value([1,1,1]), 
            potential_shadings))
        def contained_in_shadings(block):
            for shading in shadings_exclude_white:
                if shading.contains(block, threshold=constants.FACTOR_MOST): return True
            return False

        # store text lines in a potential table
        res = []           # type: list[Lines]
        table_lines = []   # type: list[Line]
        def close_table():
            if not table_lines: return
            res.append(Lines(table_lines))
            table_lines.clear()
        
        
        # check row by row 
        ref_pos = rows[0].bbox.y1
        cell_layout = isinstance(self.parent, Cell)
        for row in rows:

            bbox = row.bbox

            # flow layout or not?
            if not row.is_flow_layout(line_separate_threshold, cell_layout=cell_layout): 
                table_lines.extend([sub_line(block) for block in row])

            else:
                close_table()

            # contained in shading or not?
            for block in row:
                if contained_in_shadings(block): table_lines.append(sub_line(block))
            
            # close table if significant vertical distance 
            if bbox.y0-ref_pos>=50: close_table()
            
            # update reference pos
            ref_pos = bbox.y1

        # don't forget last table
        close_table()

        return res


    def parse_block(self, max_line_spacing_ratio:float, line_break_free_space_ratio:float, new_paragraph_free_space_ratio:float):
        '''Group lines into text block.'''
        # sort in normal reading order
        self.sort_in_reading_order_plus()

        # join lines with similar properties, e.g. spacing, together into text block
        blocks = self._join_lines_vertically(max_line_spacing_ratio)

        # split text block by checking text
        blocks = self._split_text_block_vertically(blocks,
            line_break_free_space_ratio, 
            new_paragraph_free_space_ratio)
        
        self.reset(blocks)


    def parse_text_format(self, rects, delete_end_line_hyphen:bool):
        '''Parse text format with style represented by stroke/fill shapes.
        
        Args:
            rects (Shapes): Potential styles applied on blocks.
            delete_end_line_hyphen (bool): delete hyphen at the end of a line if True.
        '''
        # parse text block style one by one
        for block in filter(lambda e: e.is_text_block, self._instances): 
            block.parse_text_format(rects)

            # adjust word at the end of each line
            block.lines.adjust_last_word(delete_end_line_hyphen)

    
    def parse_spacing(self, *args):
        '''Calculate external and internal space for text blocks:

        - vertical distance between blocks, i.e. paragraph before/after spacing
        - horizontal distance to left/right border, i.e. paragraph left/right indent
        - vertical distance between lines, i.e. paragraph line spacing
        '''
        if not self._instances: return
        self._parse_block_horizontal_spacing(*args)
        self._parse_block_vertical_spacing()
        self._parse_line_spacing()


    def make_docx(self, doc):
        '''Create page based on parsed block structure. 
        
        Args:
            doc (Document, _Cell): The container to make docx content.
        '''
        def make_table(table_block, pre_table):
            # create dummy paragraph if table before space is set
            # - a minimum line height of paragraph is 0.7pt, so ignore before space if less than this value
            # - but tow adjacent tables will be combined automatically, so adding a minimum dummy paragraph is required
            if table_block.before_space>=constants.MIN_LINE_SPACING or pre_table:
                h = lower_round(table_block.before_space, 1) # round(x,1), but to lower bound
                p = doc.add_paragraph()
                reset_paragraph_format(p, line_spacing=Pt(h))

            # new table            
            table = doc.add_table(rows=table_block.num_rows, cols=table_block.num_cols)
            table.autofit = False
            table.allow_autofit  = False
            table_block.make_docx(table)

        pre_table = False
        cell_layout = isinstance(self.parent, Cell)
        for block in self._instances:
            # make paragraphs
            if block.is_text_image_block:                
                # new paragraph
                p = doc.add_paragraph()
                block.make_docx(p)

                pre_table = False # mark block type
            
            # make table
            elif block.is_table_block:
                make_table(block, pre_table)
                pre_table = True # mark block type

                # NOTE: within a cell, there is always an empty paragraph after table,
                # so, delete it right here.
                # https://github.com/dothinking/pdf2docx/issues/76 
                if cell_layout:
                    delete_paragraph(doc.paragraphs[-1])
       
        # NOTE: If a table is at the end of a page, a new paragraph will be automatically 
        # added by the rending engine, e.g. MS Word, which resulting in an unexpected
        # page break. The solution is to never put a table at the end of a page, so add
        # an empty paragraph and reset its format, particularly line spacing, when a table
        # is created.
        for block in self._instances[::-1]:
            # ignore float image block
            if block.is_float_image_block: continue

            # nothing to do if not end with table block
            if not block.is_table_block: break

            # otherwise, add a small paragraph
            p = doc.add_paragraph()
            reset_paragraph_format(p, Pt(constants.MIN_LINE_SPACING)) # a small line height

  
    def plot(self, page):
        '''Plot blocks in PDF page for debug purpose.'''
        for block in self._instances: block.plot(page)                


    # ----------------------------------------------------------------------------------
    # internal methods
    # ----------------------------------------------------------------------------------
    def _identify_floating_images(self, float_image_ignorable_gap:float):
        '''Identify floating image lines and convert to ImageBlock.'''
        # group lines by connectivity
        groups = self.group_by_connectivity(dx=-float_image_ignorable_gap, dy=-float_image_ignorable_gap)
        
        # identify floating images
        for group in filter(lambda group: len(group)>1, groups):
            for line in filter(lambda line: line.image_spans, group):
                float_image = ImageBlock().from_image(line.spans[0])
                float_image.set_float_image_block()
                self._floating_image_blocks.append(float_image)

                # remove the original image line from flow layout by setting empty bbox
                line.update_bbox((0,0,0,0))

        return self

    def _remove_overlapped_lines(self, line_overlap_threshold:float):
        '''Delete overlapped lines. 
        NOTE: Don't run this method until floating images are excluded.
        '''
        # group lines by overlap
        fun = lambda a, b: a.get_main_bbox(b, threshold=line_overlap_threshold)
        groups = self.group(fun)
        
        # delete overlapped lines
        for group in filter(lambda group: len(group)>1, groups):
            # keep only the line with largest area
            sorted_lines = sorted(group, key=lambda line: line.bbox.get_area())
            for line in sorted_lines[:-1]:
                logging.warning('Ignore Line "%s" due to overlap', line.text)
                line.update_bbox((0,0,0,0))

        return self



    @staticmethod
    def _assign_block_to_tables(block, tables:list, blocks_in_tables:list, blocks:list):
        '''Assign block (line or table block) to contained table region ``blocks_in_tables``,
        or out-of-table ``blocks``.'''
        for table, blocks_in_table in zip(tables, blocks_in_tables):
            # fully contained in a certain table with margin
            if table.contains(block, threshold=constants.FACTOR_MAJOR):
                blocks_in_table.append(block)
                break
            
            # not possible in current table, then check next table
            elif not table.bbox.intersects(block.bbox): 
                continue
        
        # Now, this block is out of all table regions
        else:
            blocks.append(block)


    def _join_lines_vertically(self, max_line_spacing_ratio:float):
        '''Create text blocks by merge lines with same properties (spacing, font, size) in 
        vertical direction. At this moment, the block instance is either Line or TableBlock.
        '''
        idx0, idx1 = (1, 3) if self.is_horizontal_text else (0, 2)
        def get_v_bdy(block):
            '''Coordinates of block top and bottom boundaries.'''
            bbox = block.bbox
            return bbox[idx0], bbox[idx1]

        def vertical_distance(block1, block2):
            u0, u1 = get_v_bdy(block1)
            v0, v1 = get_v_bdy(block2)
            return round(v0-u1, 2)
        
        def line_height(line:Line):
            '''The height of line span with most characters.'''
            span = max(line.spans, key=lambda s: len(s.text))
            h = span.bbox.height if line.is_horizontal_text else span.bbox.width
            return round(h, 2)

        def common_vertical_spacing():
            '''Vertical distance with most frequency: a reference of line spacing.'''        
            ref0, ref1 = get_v_bdy(self._instances[0])
            distances = []
            for block in self._instances[1:]:
                y0, y1 = get_v_bdy(block)
                distances.append(round(y0-ref1, 2))
                ref0, ref1 = y0, y1        
            return max(distances, key=distances.count) if distances else 0.0

        # create text block based on lines
        blocks = [] # type: list[TextBlock]
        lines = []  # type: list[Line]
        def close_text_block():
            if not lines: return
            block = TextBlock()
            block.add(lines)
            blocks.append(block)
            lines.clear()

        # check line by line
        ref_dis = common_vertical_spacing()
        for block in self._instances:
            
            # if current is a table block:
            # - finish previous text block; and
            # - add this table block directly 
            if isinstance(block, TableBlock):
                close_text_block()
                blocks.append(block)
            
            # check two adjacent text lines
            else:
                ref_line = lines[-1] if lines else None

                # first line or in same row with previous line: needn't to create new text block
                if not ref_line or ref_line.in_same_row(block):
                    start_new_block = False
                
                # image line: create new text block
                elif block.image_spans or ref_line.image_spans:
                    start_new_block = True
                
                # lower than common line spacing: needn't to create new text block
                elif vertical_distance(ref_line, block)<=ref_dis+1.0 and \
                    ref_dis<=max_line_spacing_ratio*line_height(ref_line):
                    start_new_block = False
                
                else:
                    start_new_block = True
                
                if start_new_block: close_text_block()
                lines.append(block)

        # don't forget last group
        close_text_block()
       
        return blocks


    @staticmethod
    def _split_text_block_vertically(instances:list, line_break_free_space_ratio:float, new_paragraph_free_space_ratio:float):
        '''Split text block into separate paragraph based on punctuation of sentence.

        .. note::
            Considered only normal reading direction, from left to right, from top
            to bottom.
        '''
        blocks = [] # type: list[TextBlock]
        for block in instances:

            # add block if this isn't a text block
            if not block.is_text_block: 
                blocks.append(block)
                continue
            
            # add split blocks if necessary
            lines_list = block.lines.split_vertically_by_text(line_break_free_space_ratio, 
                                                                new_paragraph_free_space_ratio)
            if len(lines_list)==1:
                blocks.append(block)
            else:
                for lines in lines_list:
                    text_block = TextBlock()
                    text_block.add(lines)
                    blocks.append(text_block)

        return blocks


    def _parse_block_horizontal_spacing(self, *args):
        '''Calculate external horizontal space for text blocks, i.e. alignment mode and left spacing 
        for paragraph in docx:
        
            - horizontal block -> take left boundary as reference
            - vertical block   -> take bottom boundary as reference
        '''
        # bbox of blocks
        # - page level, e.g. blocks in top layout
        # - table level, e.g. blocks in table cell
        bbox = self.parent.working_bbox

        for block in self._instances:
            block.parse_horizontal_spacing(bbox, *args)


    def _parse_block_vertical_spacing(self):
        '''Calculate external vertical space for text blocks, i.e. before/after space in docx.
        
        The vertical spacing is determined by the vertical distance to previous block.
        For the first block, the reference position is top margin.

        It's easy to set before-space or after-space for a paragraph with ``python-docx``,
        so, if current block is a paragraph, set before-space for it; if current block is 
        not a paragraph, e.g. a table, set after-space for previous block (generally, 
        previous block should be a paragraph).
        '''
        # bbox of blocks
        # - page level, e.g. blocks in top layout
        # - table level, e.g. blocks in table cell
        bbox = self.parent.working_bbox

        # check text direction for vertical space calculation:
        # - normal reading direction (from left to right)    -> the reference boundary is top border, i.e. bbox[1].
        # - vertical text direction, e.g. from bottom to top -> left border bbox[0] is the reference
        idx = 1 if self.is_horizontal_text else 0

        ref_block = self._instances[0]
        ref_pos = bbox[idx]

        for block in self._instances:

            # NOTE: the table bbox is counted on center-line of outer borders, so a half of top border
            # size should be excluded from the calculated vertical spacing
            if block.is_table_block:
                dw = block[0][0].border_width[0] / 2.0 # use top border of the first cell
            
            else:
                dw = 0.0

            start_pos = block.bbox[idx] - dw
            para_space = start_pos-ref_pos

            # modify vertical space in case the block is out of bottom boundary
            dy = max(block.bbox[idx+2]-bbox[idx+2], 0.0)
            para_space -= dy
            para_space = max(para_space, 0.0) # ignore negative value

            # ref to current (paragraph): set before-space for paragraph
            if block.is_text_block:
                # spacing before this paragraph
                block.before_space = para_space

            # if ref to current (image): set before-space for paragraph
            elif block.is_inline_image_block:
                block.before_space = para_space

            # ref (paragraph/image) to current: set after-space for ref paragraph        
            elif ref_block.is_text_block or ref_block.is_inline_image_block:
                ref_block.after_space = para_space

            # situation with very low probability, e.g. ref (table) to current (table)
            # we can't set before space for table in docx, but the tricky way is to
            # create an empty paragraph and set paragraph line spacing and before space
            else:
                # let para_space>=1 Pt to accommodate the dummy paragraph if not the first block
                block.before_space = max(para_space, int(block!=self._instances[0])*constants.MINOR_DIST)

            # update reference block        
            ref_block = block
            ref_pos = ref_block.bbox[idx+2] + dw # assume same bottom border with top one

        # NOTE: when a table is at the end of a page, a dummy paragraph with a small line spacing 
        # is added after this table, to avoid unexpected page break. Accordingly, this extra spacing 
        # must be remove in other place, especially the page space is run out.
        # Here, reduce the last row of table.
        block = self._instances[-1]
        if block.is_table_block: block[-1].height -= constants.MINOR_DIST


    def _parse_line_spacing(self):
        '''Calculate internal vertical space for text blocks, i.e. paragraph line spacing in docx.

        .. note::
            Run parsing block vertical spacing in advance.
        '''
        def is_exact_line_spacing(block):
            '''check line spacing type based on parsed font metrics of text span:
            exact line spacing if no standard line height is extracted
            '''
            for line in block.lines:
                absent_line_heights = list(not span.is_valid_line_height \
                                        for span in line.spans if isinstance(span, TextSpan))
                if any(absent_line_heights): return True
            return False

        for block in self._instances:
            if not block.is_text_block: continue
            
            if is_exact_line_spacing(block):
                block.line_space_type = 0
                block.parse_exact_line_spacing()
            else:
                block.parse_relative_line_spacing()



# -*- coding: utf-8 -*-

'''A group of ``Shape`` instances.
'''

from .Shape import Shape, Stroke, Fill, Hyperlink
from ..common.share import RectType
from ..common.Collection import Collection, ElementCollection
from ..common import share
from ..common import constants


class Shapes(ElementCollection):
    ''' A collection of ``Shape`` instances: ``Stroke`` or ``Fill``.'''

    def restore(self, raws:list):
        '''Clean current instances and restore them from source dicts.'''
        self.reset()
        # Distinguish specified type by key like `start`, `end` and `uri`.
        for raw in raws:
            if 'start' in raw:
                shape = Stroke(raw)
            elif 'uri' in raw:
                shape = Hyperlink(raw)
            else:
                shape = Fill(raw)
            # add to list
            self.append(shape)
        
        return self


    def _update_bbox(self, shape:Shape):
        ''' override. Do nothing.'''
        pass


    @property
    def strokes(self):
        ''' Stroke Shapes, including table border, text underline and strike-through. 
            Cache it once calculated since it doesn't change generally.
        '''
        instances = list(filter(
            lambda shape: isinstance(shape, Stroke), self._instances))
        return Shapes(instances)


    @property
    def fillings(self):
        ''' Fill Shapes, including cell shading and highlight. 
            Cache it once calculated since it doesn't change generally.
        '''
        # white bg-color is by default, so ignore those fillings
        instances = list(filter(
            lambda shape: isinstance(shape, Fill) and \
                shape.color != share.rgb_value((1,1,1)), self._instances))
        return Shapes(instances)


    @property
    def hyperlinks(self):
        ''' Hyperlink Shapes.'''
        instances = list(filter(
            lambda shape: isinstance(shape, Hyperlink), self._instances))
        return Shapes(instances)


    @property
    def table_strokes(self):
        '''Potential table borders.'''
        instances = list(filter(
            lambda shape: shape.has_potential_type(RectType.BORDER), self._instances))
        return ElementCollection(instances)

    
    @property
    def table_fillings(self):
        '''Potential table shadings.'''
        instances = list(filter(
            lambda shape: shape.has_potential_type(RectType.SHADING), self._instances))
        return ElementCollection(instances)
    
    @property
    def text_style_shapes(self):
        '''Potential text style based shapes, e.g. underline, strike-through, highlight and hyperlink.'''
        f = lambda shape: shape.has_potential_type(RectType.HIGHLIGHT) or \
                            shape.has_potential_type(RectType.UNDERLINE) or \
                            shape.has_potential_type(RectType.STRIKE) or \
                            shape.has_potential_type(RectType.HYPERLINK)
        instances = set(filter(f, self._instances))
        return ElementCollection(instances)


    def clean_up(self, max_border_width:float, shape_min_dimension:float):
        """Clean rectangles.

        * Delete shapes out of page.
        * Delete small shapes (either width or height).
        * Merge shapes with same filling color.
        * Detect semantic type.

        Args:
            max_border_width (float): The max border width.
            shape_min_dimension (float): Ignore shape if both width and height is lower than this value.
        """
        if not self._instances: return

        # remove small shapes or shapes out of page
        page_bbox = self.parent.bbox
        f = lambda shape: shape.bbox.intersects(page_bbox) and \
                        max(shape.bbox.width, shape.bbox.height)>=shape_min_dimension
        cleaned_shapes = list(filter(f, self._instances)) # type: list[Shape]

        # merge normal shapes if same filling color
        merged_shapes = self._merge_shapes(cleaned_shapes)
                
        # convert Fill instance to Stroke if looks like stroke
        shapes = []
        for shape in merged_shapes:
            if isinstance(shape, Fill):
                stroke = shape.to_stroke(max_border_width)
                shapes.append(stroke if stroke else shape)
            else:
                shapes.append(shape)
        self.reset(shapes)

        # detect semantic type
        self._parse_semantic_type()
    

    def assign_to_tables(self, tables:list):
        """Add Shape to associated cells of given tables.

        Args:
            tables (list): A list of TableBlock instances.
        """
        if not tables: return

        # assign shapes to table region        
        shapes_in_tables = [[] for _ in tables] # type: list[list[Shape]]
        shapes = []   # type: list[Shape]
        for shape in self._instances:
            # exclude explicit table borders which belongs to current layout
            if shape.equal_to_type(RectType.BORDER) or shape.equal_to_type(RectType.SHADING):
                shapes.append(shape)
                continue

            for table, shapes_in_table in zip(tables, shapes_in_tables):
                # fully contained in one table
                if table.bbox.contains(shape.bbox):
                    shapes_in_table.append(shape)
                    break

                # not possible in current table, then check next table
                elif not table.bbox.intersects(shape.bbox):
                    continue
            
            # Now, this shape belongs to previous layout
            else:
                shapes.append(shape)

        # assign shapes to associated cells
        for table, shapes_in_table in zip(tables, shapes_in_tables):
            # no contents for this table
            if not shapes_in_table: continue
            table.assign_shapes(shapes_in_table)

        self.reset(shapes)


    def plot(self, page):
        '''Plot shapes for debug purpose. Different colors are used to display the shapes in detected 
        semantic types, e.g. yellow for text based shape (stroke, underline and highlight). Due to 
        overlaps between Stroke and Fill related groups, some shapes are plot twice.
        
        Args:
            page (fitz.Page): pdf page.
        '''
        # Table based shapes
        # - table shading
        color = (152/255, 251/255, 152/255)
        for shape in self.table_fillings: shape.plot(page, color)

        # - table borders
        color = (0, 0, 0)
        for shape in self.table_strokes: shape.plot(page, color)

        # Text based shapes
        # - underline
        # - strike-through
        # - highlight
        # - hyperlink
        color = (1, 1, 0)
        for shape in self.text_style_shapes: shape.plot(page, color)


    @staticmethod
    def _merge_shapes(shapes):
        '''Merge shapes if same filling color. Note the merged bbox must match source shapes
        as more as possible.'''
        # shapes excluding hyperlink first
        normal_shapes = list(filter(
            lambda shape: not shape.is_determined, shapes))
        
        # group by color and connectivity (with margin considered)
        f = lambda a, b: \
            a.color==b.color and a.bbox.intersects(b.get_expand_bbox(constants.TINY_DIST))
        groups = Collection(normal_shapes).group(f)

        merged_shapes = []
        for group in groups:
            merged_area = group.bbox.get_area()
            sum_area = sum(shape.bbox.get_area() for shape in group)
            if sum_area/merged_area >= constants.FACTOR_ALMOST:
                merged_shapes.append(group[0].update_bbox(group.bbox))
            else:
                merged_shapes.extend(group)
        
        # add hyperlinks back
        hyperlinks = filter(lambda shape: shape.equal_to_type(RectType.HYPERLINK), shapes)
        merged_shapes.extend(hyperlinks)
        
        return merged_shapes


    def _parse_semantic_type(self):
        ''' Detect shape type based on the position to text blocks. 

        .. note::
            Stroke shapes are grouped on connectivity to each other, but in some cases, 
            the gap between borders and underlines/strikes are very close, which leads
            to an incorrect table structure. So, it's required to distinguish them in
            advance, though we needn't to ensure 100% accuracy. They are finally determined 
            when parsing table structure and text format.
        '''
        # blocks in page (the original blocks without any further processing)
        blocks = self._parent.blocks
        blocks.sort_in_reading_order()

        # check positions between shapes and text blocks
        for shape in self._instances:
            shape.parse_semantic_type(blocks)




# -*- coding: utf-8 -*-

'''Text Span object based on PDF raw dict extracted with ``PyMuPDF``.

Data structure for Span refer to 
this `link <https://pymupdf.readthedocs.io/en/latest/textpage.html>`_::

    {
        # raw dict
        ---------------------------
        'bbox': (x0,y0,x1,y1),
        'color': sRGB
        'font': fontname,
        'size': fontsize,
        'flags': fontflags,
        'chars': [ chars ],

        # added dict
        ----------------------------
        'text': text,
        'style': [
            {
                'type': int,
                'color': int,
                'uri': str    # for hyperlink
            },
            ...
        ]
    }
'''

import fitz
from docx.shared import Pt, RGBColor
from docx.oxml.ns import qn
from .Char import Char
from ..common.Element import Element
from ..common.share import RectType
from ..common import constants
from ..common import share, docx
from ..shape.Shape import Shape


class TextSpan(Element):
    '''Object representing text span.'''
    def __init__(self, raw:dict=None):
        raw = raw or {}
        self.color = raw.get('color', 0)
        self.flags = raw.get('flags', 0)

        # filter empty chars
        chars = [Char(c) for c in raw.get('chars', [])] # type: list[Char]
        self.chars = [char for char in chars if char.c!='']
        self._text = raw.get('text', '') # not an original key from PyMuPDF

        # font metrics
        # line_height is the standard single line height used in relative line spacing,
        # while exact line spacing is used when line_height==-1 by default.
        self.font = raw.get('font', '')
        self.size = raw.get('size', 12.0)
        self.ascender = raw.get('ascender', 1.0)
        self.descender = raw.get('descender', 0.0)
        self.line_height = raw.get('line_height', -1)  # not an original key

        # introduced attributes
        # a list of dict: { 'type': int, 'color': int }
        self.style = raw.get('style', [])

        # char spacing between adjacent two chars -> pdf operador Tc
        # positive to expand space, otherwise condense
        # just an attribute placeholder: not used yet
        self.char_spacing = raw.get('char_spacing', 0.0)
        
        # init text span element
        super().__init__(raw)

        # in rare case, the font is unamed, so change font and update bbox accordingly
        if 'UNNAMED' in self.font.upper():
            self._change_font_and_update_bbox(constants.DEFAULT_FONT_NAME)


    @property
    def text(self):
        '''Get span text. Note joining chars is in a higher priority.'''
        return ''.join([char.c for char in self.chars]) if self.chars else self._text

    @text.setter
    def text(self, value):
        '''Set span text directly in case no chars are stores, e.g. restored from json.'''
        self._text = value
    
    def cal_bbox(self):
        '''Calculate bbox based on contained instances.'''
        bbox = fitz.Rect()
        for char in self.chars: bbox |= char.bbox
        return bbox

    @property
    def is_valid_line_height(self): return self.line_height!=-1


    def _change_font_and_update_bbox(self, font_name:str):
        '''Set new font, and update font size, span/char bbox accordingly.

        It's generally used for span with unnamed fonts. 
        See this `issue <https://github.com/pymupdf/PyMuPDF/issues/642>`_.        

        In corner case, where the PDF file containing unnamed and not embedded fonts, the span bbox
        extracted from ``PyMuPDF`` is not correct. ``PyMuPDF`` provides feature to replace these 
        unnamed fonts with specified fonts, then extract correct bbox from the updated PDF. Since we 
        care less about the original PDF itself but its layout, the idea here is to set a default font 
        for text spans with unnamed fonts, and estimate the updated bbox with method from 
        ``fitz.TextWriter``.

        Args:
            font_name (str): Font name.
        '''
        # set new font property
        self.font = font_name

        # compute text length under new font with that size
        font = fitz.Font(font_name)
        new_length = font.text_length(self.text, fontsize=self.size)
        if new_length > self.bbox.width:
            self.size *= self.bbox.width / new_length

        # estimate occupied rect when added with TextWriter
        x0, y0, x1, y1 = self.bbox
        tw = fitz.TextWriter((0, 0, x1, y1))
        rect, _ = tw.append(
            self.chars[0].origin, # the bottom left point of the first character
            self.text,
            font=font,
            fontsize=self.size
        )

        # update span bbox
        # - x-direction: use original horizontal range
        # - y-direction: centerline defined by estimated vertical range, and height by font size
        buff = (rect.height-self.size)/2.0
        y0 = rect.y0 + buff
        y1 = rect.y1 - buff
        self.update_bbox((x0, y0, x1, y1))

        # update contained char bbox
        for char in self.chars:
            x0, _, x1, _ = char.bbox
            char.update_bbox((x0, y0, x1, y1))


    def add(self, char:Char):
        '''Add char and update bbox accordingly.'''
        self.chars.append(char)
        self.union_bbox(char)

    
    def lstrip(self):
        '''Remove blanks at the left side, but keep one blank.'''
        original_text = self.text
        if not original_text.startswith(' '*2): return False

        # keep one blank
        num_blanks = len(original_text) - len(original_text.lstrip())
        self.chars = self.chars[num_blanks-1:]
        self.update_bbox(rect=self.cal_bbox())
        return True
    

    def rstrip(self):
        '''Remove blanks at the right side, but keep one blank.'''
        original_text = self.text
        if not original_text.endswith(' '*2): return False

        # keep one blank
        num_blanks = len(original_text) - len(original_text.rstrip())
        self.chars = self.chars[:1-num_blanks]
        self.update_bbox(rect=self.cal_bbox())
        return True


    def store(self):
        res = super().store()
        res.update({
            'color': self.color,
            'font': self.font,
            'size': self.size,
            'line_height': self.line_height, 
            'flags': self.flags,
            'text': self.text,
            'style': self.style,
            'char_spacing': self.char_spacing
        }) # not storing chars for space saving
        return res


    def plot(self, page, color:tuple): super().plot(page, stroke=color, fill=color, width=0)


    def split(self, rect:Shape, horizontal:bool=True):
        """Split span with the intersection: span-intersection-span.

        Args:
            rect (Shape): Target shape to split this text span.
            horizontal (bool, optional): Text direction. Defaults to True.

        Returns:
            list: Split text spans.
        """        
        # any intersection in this span?
        # NOTE: didn't consider the case that an underline is out of a span
        intsec = rect.bbox & self.bbox

        # no, then add this span as it is
        # Note the case bool(intsec)=True but intsec.get_area()=0
        if intsec.is_empty: return [self]
        

        # yes, then split spans:
        # - add new style to the intersection part
        # - keep the original style for the rest
        split_spans = [] # type: list[TextSpan]

        # expand the intersection area, e.g. for strike through line,
        # the intersection is a `line`, i.e. a rectangle with very small height,
        # so expand the height direction to span height
        if horizontal:
            intsec.y0 = self.bbox.y0
            intsec.y1 = self.bbox.y1
        else:
            intsec.x0 = self.bbox.x0
            intsec.x1 = self.bbox.x1

        # calculate chars in the format rectangle
        # combine an index with enumerate(), so the second element is the char
        f = lambda items: items[1].contained_in_rect(rect, horizontal)
        index_chars = list(filter(f, enumerate(self.chars)))

        # then we get target chars in a sequence
        pos = index_chars[0][0] if index_chars else -1 # start index -1 if nothing found
        length = len(index_chars)
        pos_end = max(pos+length, 0) # max() is used in case: pos=-1, length=0

        # split span with the intersection: span-intersection-span
        # 
        # left part if exists
        if pos > 0:
            if horizontal:
                bbox = (self.bbox.x0, self.bbox.y0, intsec.x0, self.bbox.y1)
            else:
                bbox = (self.bbox.x0, intsec.y1, self.bbox.x1, self.bbox.y1)
            split_span = self.copy().update_bbox(bbox)
            split_span.chars = self.chars[0:pos]
            split_spans.append(split_span)

        # middle intersection part if exists
        if length > 0:
            bbox = (intsec.x0, intsec.y0, intsec.x1, intsec.y1)
            split_span = self.copy().update_bbox(bbox)
            split_span.chars = self.chars[pos:pos_end]            
            split_span._parse_text_format(rect, horizontal)  # update style
            split_spans.append(split_span)

        # right part if exists
        if pos_end < len(self.chars):
            if horizontal:
                bbox = (intsec.x1, self.bbox.y0, self.bbox.x1, self.bbox.y1)
            else:
                bbox = (self.bbox.x0, self.bbox.y0, self.bbox.x1, intsec.y0)
            split_span = self.copy().update_bbox(bbox)
            split_span.chars = self.chars[pos_end:]
            split_spans.append(split_span)

        return split_spans


    def _parse_text_format(self, rect:Shape, horizontal:bool=True):
        """Parse text style based on the position to a rect shape.

        Args:
            rect (Shape): Target rect shape representing potential text style.
            horizontal (bool, optional): Horizontal text direction. Defaults to True.

        Returns:
            bool: Parsed text style successfully or not.
        """

        # Skip table border/shading
        if rect.equal_to_type(RectType.BORDER) or rect.equal_to_type(RectType.SHADING):
            return False
        
        # set hyperlink
        elif rect.equal_to_type(RectType.HYPERLINK):
            self.style.append({
                'type': rect.type,
                'color': rect.color,
                'uri': rect.uri
            })
            return True

        # considering text direction
        idx = 1 if horizontal else 0

        # recognize text format based on rect and the span it applying to
        # region height
        h_rect = rect.bbox[idx+2] - rect.bbox[idx]
        h_span = self.bbox[idx+2] - self.bbox[idx]

        # distance to span bottom border
        d = abs(self.bbox[idx+2] - rect.bbox[idx])

        # highlight: both the rect height and overlap must be large enough
        if h_rect >= 0.5*h_span:
            # In general, highlight color isn't white
            if rect.color != share.rgb_value((1,1,1)) and self.get_main_bbox(rect, constants.FACTOR_MAJOR): 
                rect.type = RectType.HIGHLIGHT
    
        # near to bottom of span? yes, underline
        elif d <= 0.25*h_span:
            rect.type = RectType.UNDERLINE

        # near to center of span? yes, strike-through-line
        elif 0.35*h_span < d < 0.75*h_span:
            rect.type = RectType.STRIKE

        # check rect type again
        if not rect.is_determined: return False

        style =  {
            'type': rect.type,
            'color': rect.color
        }
        self.style.append(style)

        return True


    def intersects(self, rect):
        '''Create new TextSpan object with chars contained in given bbox.
        
        Args:
            rect (fitz.Rect): Target bbox.
        '''
        # add span directly if fully contained in bbox
        if rect.contains(self.bbox):
            return self.copy()

        # no intersection
        if not rect.intersects(self.bbox):
            return TextSpan()

        # further check chars in span
        span = self.copy()
        span.chars.clear()
        span.update_bbox((0.0,0.0,0.0,0.0))

        for char in self.chars:
            if char.get_main_bbox(rect, constants.FACTOR_A_HALF): # contains at least a half part
                span.chars.append(char)
                span.union_bbox(char)

        return span


    def make_docx(self, paragraph):
        '''Add text span to a docx paragraph, and set text style, e.g. font, color, underline, hyperlink, etc.

        .. note::
            Hyperlink and its style is parsed separately from pdf. For instance, regarding a general hyperlink with an
            underline, the text and uri is parsed as hyperlink itself, while the underline is treated as a normal text
            style.
        '''
        # Create hyperlink in particular, otherwise add a run directly
        for style in self.style:
            if style['type']==RectType.HYPERLINK.value and self.text.strip():
                docx_run = docx.add_hyperlink(paragraph, style['uri'], self.text)
                break
        else:
            docx_run = paragraph.add_run(self.text)
        
        # set text style, e.g. font, underline and highlight
        self._set_text_format(docx_run)

        # set charters spacing
        if self.char_spacing: 
            docx.set_char_spacing(docx_run, self.char_spacing)


    def _set_text_format(self, docx_run):
        '''Set text format for ``python-docx.run`` object.'''
        # set style
        # https://python-docx.readthedocs.io/en/latest/api/text.html#docx.text.run.Font

        # basic font style
        # line['flags'] is an integer, encoding bool of font properties:
        # bit 0: superscripted (2^0)
        # bit 1: italic (2^1)
        # bit 2: serifed (2^2)
        # bit 3: monospaced (2^3)
        # bit 4: bold (2^4)
        docx_run.superscript = bool(self.flags & 2**0)
        docx_run.italic = bool(self.flags & 2**1)
        docx_run.bold = bool(self.flags & 2**4)

        # font name
        font_name = self.font
        docx_run.font.name = font_name
        docx_run._element.rPr.rFonts.set(qn('w:eastAsia'), font_name) # set font for chinese characters
        docx_run.font.color.rgb = RGBColor(*share.rgb_component(self.color))

        # font size
        # NOTE: only x.0 and x.5 is accepted in docx, so set character scaling accordingly
        # if the font size doesn't meet this condition.
        font_size = round(self.size*2)/2.0
        docx_run.font.size = Pt(font_size)

        # adjust by set scaling
        scale = self.size / (font_size or self.size or 1)
        if abs(scale-1.0)>=0.01:
            docx.set_char_scaling(docx_run, scale)
        
        # font style parsed from PDF rectangles: 
        # e.g. highlight, underline, strike-through-line
        for style in self.style:
            
            t = style['type']
            # Built-in method is provided to set highlight in python-docx, but supports only limited colors;
            # so, set character shading instead if out of highlight color scope
            if t==RectType.HIGHLIGHT.value:
                docx.set_char_shading(docx_run, style['color'])

            # underline set with built-in method `font.underline` has a same color with text.
            # so, try to set a different color with xml if necessary
            elif t==RectType.UNDERLINE.value:
                if self.color==style['color']:
                    docx_run.font.underline = True
                else:
                    docx.set_char_underline(docx_run, style['color'])
            
            # same color with text for strike line
            elif t==RectType.STRIKE.value:
                docx_run.font.strike = True